# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import tempfile
from os import path, mkdir
from shutil import copyfile, rmtree
from datetime import datetime
from werkzeug.datastructures import FileStorage
from mock import patch
import module_build_service.utils
import module_build_service.scm
from module_build_service import models, conf
from module_build_service.errors import ProgrammingError, ValidationError, UnprocessableEntity
from tests import (
    reuse_component_init_data, db, reuse_shared_userspace_init_data, clean_database, init_data,
    scheduler_init_data)
import mock
import koji
import pytest
import module_build_service.scheduler.handlers.components
from module_build_service.builder.base import GenericBuilder
from module_build_service.builder.KojiModuleBuilder import KojiModuleBuilder
from module_build_service import glib, Modulemd
from tests import app

BASE_DIR = path.abspath(path.dirname(__file__))


class FakeSCM(object):
    def __init__(self, mocked_scm, name, mmd_filename, commit=None):
        self.mocked_scm = mocked_scm
        self.name = name
        self.commit = commit
        self.mmd_filename = mmd_filename
        self.sourcedir = None

        self.mocked_scm.return_value.checkout = self.checkout
        self.mocked_scm.return_value.name = self.name
        self.mocked_scm.return_value.branch = 'master'
        self.mocked_scm.return_value.get_latest = self.get_latest
        self.mocked_scm.return_value.commit = self.commit
        self.mocked_scm.return_value.repository_root = "git://pkgs.stg.fedoraproject.org/modules/"
        self.mocked_scm.return_value.sourcedir = self.sourcedir
        self.mocked_scm.return_value.get_module_yaml = self.get_module_yaml

    def checkout(self, temp_dir):
        self.sourcedir = path.join(temp_dir, self.name)
        mkdir(self.sourcedir)
        base_dir = path.abspath(path.dirname(__file__))
        copyfile(path.join(base_dir, '..', 'staged_data', self.mmd_filename),
                 self.get_module_yaml())

        return self.sourcedir

    def get_latest(self, ref='master'):
        return self.commit if self.commit else ref

    def get_module_yaml(self):
        return path.join(self.sourcedir, self.name + ".yaml")


class TestUtilsComponentReuse:

    def setup_method(self, test_method):
        reuse_component_init_data()

    def teardown_method(self, test_method):
        clean_database()

    @pytest.mark.parametrize('changed_component', [
        'perl-List-Compare', 'perl-Tangerine', 'tangerine', None
    ])
    def test_get_reusable_component_different_component(self, changed_component):
        second_module_build = models.ModuleBuild.query.filter_by(id=3).one()
        if changed_component:
            mmd = second_module_build.mmd()
            mmd.get_rpm_components()['tangerine'].set_ref(
                '00ea1da4192a2030f9ae023de3b3143ed647bbab')
            second_module_build.modulemd = mmd.dumps()
            second_module_changed_component = models.ComponentBuild.query.filter_by(
                package=changed_component, module_id=3).one()
            second_module_changed_component.ref = '00ea1da4192a2030f9ae023de3b3143ed647bbab'
            db.session.add(second_module_changed_component)
            db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        tangerine_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')

        if changed_component == 'perl-List-Compare':
            # perl-Tangerine can be reused even though a component in its batch has changed
            assert plc_rv is None
            assert pt_rv.package == 'perl-Tangerine'
            assert tangerine_rv is None
        elif changed_component == 'perl-Tangerine':
            # perl-List-Compare can be reused even though a component in its batch has changed
            assert plc_rv.package == 'perl-List-Compare'
            assert pt_rv is None
            assert tangerine_rv is None
        elif changed_component == 'tangerine':
            # perl-List-Compare and perl-Tangerine can be reused since they are in an earlier
            # buildorder than tangerine
            assert plc_rv.package == 'perl-List-Compare'
            assert pt_rv.package == 'perl-Tangerine'
            assert tangerine_rv is None
        elif changed_component is None:
            # Nothing has changed so everthing can be used
            assert plc_rv.package == 'perl-List-Compare'
            assert pt_rv.package == 'perl-Tangerine'
            assert tangerine_rv.package == 'tangerine'

    def test_get_reusable_component_different_rpm_macros(self):
        second_module_build = models.ModuleBuild.query.filter_by(id=3).one()
        mmd = second_module_build.mmd()
        mmd.set_rpm_buildopts({'macros': '%my_macro 1'})
        second_module_build.modulemd = mmd.dumps()
        db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        assert plc_rv is None

        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        assert pt_rv is None

    @pytest.mark.parametrize('set_current_arch', [True, False])
    @pytest.mark.parametrize('set_database_arch', [True, False])
    def test_get_reusable_component_different_arches(self, set_database_arch, set_current_arch):
        second_module_build = models.ModuleBuild.query.filter_by(id=3).one()
        if set_current_arch:       # set architecture for current build
            mmd = second_module_build.mmd()
            arches = Modulemd.SimpleSet()
            arches.set(['i686'])
            mmd.get_rpm_components()['tangerine'].set_arches(arches)
            second_module_build.modulemd = mmd.dumps()
        if set_database_arch:       # set architecture for build in database
            second_module_changed_component = models.ComponentBuild.query.filter_by(
                package='tangerine', module_id=2).one()
            mmd = second_module_changed_component.module_build.mmd()
            arches = Modulemd.SimpleSet()
            arches.set(['i686'])
            mmd.get_rpm_components()['tangerine'].set_arches(arches)
            second_module_changed_component.module_build.modulemd = mmd.dumps()
            db.session.add(second_module_changed_component)
            db.session.commit()

        tangerine = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')
        assert bool(tangerine is None) != bool(set_current_arch == set_database_arch)

    @pytest.mark.parametrize('rebuild_strategy', models.ModuleBuild.rebuild_strategies.keys())
    def test_get_reusable_component_different_buildrequires_hash(self, rebuild_strategy):
        first_module_build = models.ModuleBuild.query.filter_by(id=2).one()
        first_module_build.rebuild_strategy = rebuild_strategy
        second_module_build = models.ModuleBuild.query.filter_by(id=3).one()
        mmd = second_module_build.mmd()
        xmd = glib.from_variant_dict(mmd.get_xmd())
        xmd['mbs']['buildrequires']['platform']['ref'] = \
            'da39a3ee5e6b4b0d3255bfef95601890afd80709'
        mmd.set_xmd(glib.dict_values(xmd))
        second_module_build.modulemd = mmd.dumps()
        second_module_build.ref_build_context = '37c6c57bedf4305ef41249c1794760b5cb8fad17'
        second_module_build.rebuild_strategy = rebuild_strategy
        db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        tangerine_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')

        if rebuild_strategy == "only-changed":
            assert plc_rv is not None
            assert pt_rv is not None
            assert tangerine_rv is not None
        else:
            assert plc_rv is None
            assert pt_rv is None
            assert tangerine_rv is None

    @pytest.mark.parametrize('rebuild_strategy', models.ModuleBuild.rebuild_strategies.keys())
    def test_get_reusable_component_different_buildrequires_stream(self, rebuild_strategy):
        first_module_build = models.ModuleBuild.query.filter_by(id=2).one()
        first_module_build.rebuild_strategy = rebuild_strategy
        second_module_build = models.ModuleBuild.query.filter_by(id=3).one()
        mmd = second_module_build.mmd()
        xmd = glib.from_variant_dict(mmd.get_xmd())
        xmd['mbs']['buildrequires']['platform']['stream'] = 'different'
        mmd.set_xmd(glib.dict_values(xmd))
        second_module_build.modulemd = mmd.dumps()
        second_module_build.build_context = '37c6c57bedf4305ef41249c1794760b5cb8fad17'
        second_module_build.rebuild_strategy = rebuild_strategy
        db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        tangerine_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')

        assert plc_rv is None
        assert pt_rv is None
        assert tangerine_rv is None

    def test_get_reusable_component_different_buildrequires(self):
        second_module_build = models.ModuleBuild.query.filter_by(id=3).one()
        mmd = second_module_build.mmd()
        br_list = Modulemd.SimpleSet()
        br_list.add('master')
        mmd.get_dependencies()[0].set_buildrequires({'some_module': br_list})
        xmd = glib.from_variant_dict(mmd.get_xmd())
        xmd['mbs']['buildrequires'] = {
            'some_module': {
                'ref': 'da39a3ee5e6b4b0d3255bfef95601890afd80709',
                'stream': 'master',
                'version': '20170123140147'
            }
        }
        mmd.set_xmd(glib.dict_values(xmd))
        second_module_build.modulemd = mmd.dumps()
        second_module_build.ref_build_context = '37c6c57bedf4305ef41249c1794760b5cb8fad17'
        db.session.commit()

        plc_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-List-Compare')
        assert plc_rv is None

        pt_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'perl-Tangerine')
        assert pt_rv is None

        tangerine_rv = module_build_service.utils.get_reusable_component(
            db.session, second_module_build, 'tangerine')
        assert tangerine_rv is None

    @patch("module_build_service.utils.submit.submit_module_build")
    def test_submit_module_build_from_yaml_with_skiptests(self, mock_submit):
        """
        Tests local module build from a yaml file with the skiptests option

        Args:
            mock_submit (MagickMock): mocked function submit_module_build, which we then
                inspect if it was called with correct arguments
        """
        module_dir = tempfile.mkdtemp()
        module = models.ModuleBuild.query.filter_by(id=3).one()
        mmd = module.mmd()
        modulemd_yaml = mmd.dumps()
        modulemd_file_path = path.join(module_dir, "testmodule.yaml")

        username = "test"
        stream = "dev"

        with open(modulemd_file_path, "w") as fd:
            fd.write(modulemd_yaml)

        with open(modulemd_file_path, "r") as fd:
            handle = FileStorage(fd)
            module_build_service.utils.submit_module_build_from_yaml(username, handle,
                                                                     stream=stream, skiptests=True)
            mock_submit_args = mock_submit.call_args[0]
            username_arg = mock_submit_args[0]
            mmd_arg = mock_submit_args[2]
            assert mmd_arg.get_stream() == stream
            assert "\n\n%__spec_check_pre exit 0\n" in mmd_arg.get_rpm_buildopts()['macros']
            assert username_arg == username
        rmtree(module_dir)


class TestUtils:

    def setup_method(self, test_method):
        clean_database()

    def teardown_method(self, test_method):
        clean_database()

    def test_get_rpm_release_mse(self):
        init_data(contexts=True)
        build_one = models.ModuleBuild.query.get(2)
        build_two = models.ModuleBuild.query.get(3)
        release_one = module_build_service.utils.get_rpm_release(build_one)
        release_two = module_build_service.utils.get_rpm_release(build_two)
        assert release_one == "module+2+b8645bbb"
        assert release_two == "module+2+17e35784"

    def test_get_rpm_release_platform_stream(self):
        scheduler_init_data(1)
        build_one = models.ModuleBuild.query.get(2)
        release = module_build_service.utils.get_rpm_release(build_one)
        assert release == 'module+f28+2+814cfa39'

    @pytest.mark.parametrize('scmurl', [
        ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git'
         '?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'),
        None
    ])
    @patch('module_build_service.scm.SCM')
    def test_format_mmd(self, mocked_scm, scmurl):
        mocked_scm.return_value.commit = \
            '620ec77321b2ea7b0d67d82992dda3e1d67055b4'
        # For all the RPMs in testmodule, get_latest is called
        mocked_scm.return_value.get_latest.side_effect = [
            '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
            'fbed359411a1baa08d4a88e0d12d426fbf8f602c']
        hashes_returned = {
            'master': 'fbed359411a1baa08d4a88e0d12d426fbf8f602c',
            'f28': '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
            'f27': '5deef23acd2367d8b8d5a621a2fc568b695bc3bd'}

        def mocked_get_latest(ref="master"):
            return hashes_returned[ref]

        mocked_scm.return_value.get_latest = mocked_get_latest
        mmd = Modulemd.Module().new_from_file(
            path.join(BASE_DIR, '..', 'staged_data', 'testmodule.yaml'))
        mmd.upgrade()
        # Modify the component branches so we can identify them later on
        mmd.get_rpm_components()['perl-Tangerine'].set_ref('f28')
        mmd.get_rpm_components()['tangerine'].set_ref('f27')
        module_build_service.utils.format_mmd(mmd, scmurl)

        # Make sure that original refs are not changed.
        mmd_pkg_refs = [pkg.get_ref() for pkg in mmd.get_rpm_components().values()]
        assert set(mmd_pkg_refs) == set(hashes_returned.keys())
        br = mmd.get_dependencies()[0].get_buildrequires()
        assert br.keys() == ['platform']
        assert br.values()[0].get() == ['f28']
        xmd = {
            'mbs': {
                'commit': '',
                'rpms': {
                    'perl-List-Compare': {'ref': 'fbed359411a1baa08d4a88e0d12d426fbf8f602c'},
                    'perl-Tangerine': {'ref': '4ceea43add2366d8b8c5a622a2fb563b625b9abf'},
                    'tangerine': {'ref': '5deef23acd2367d8b8d5a621a2fc568b695bc3bd'}},
                'scmurl': ''
            }
        }
        if scmurl:
            xmd['mbs']['commit'] = '620ec77321b2ea7b0d67d82992dda3e1d67055b4'
            xmd['mbs']['scmurl'] = scmurl
        mmd_xmd = glib.from_variant_dict(mmd.get_xmd())
        assert mmd_xmd == xmd

    def test_get_reusable_component_shared_userspace_ordering(self):
        """
        For modules with lot of components per batch, there is big chance that
        the database will return them in different order than what we have for
        current `new_module`. In this case, reuse code should still be able to
        reuse the components.
        """
        reuse_shared_userspace_init_data()
        new_module = models.ModuleBuild.query.get(3)
        rv = module_build_service.utils.get_reusable_component(
            db.session, new_module, 'llvm')
        assert rv.package == 'llvm'

    def test_validate_koji_tag_wrong_tag_arg_during_programming(self):
        """ Test that we fail on a wrong param name (non-existing one) due to
        programming error. """

        @module_build_service.utils.validate_koji_tag('wrong_tag_arg')
        def validate_koji_tag_programming_error(good_tag_arg, other_arg):
            pass

        with pytest.raises(ProgrammingError):
            validate_koji_tag_programming_error('dummy', 'other_val')

    def test_validate_koji_tag_bad_tag_value(self):
        """ Test that we fail on a bad tag value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_bad_tag_value(tag_arg):
            pass

        with pytest.raises(ValidationError):
            validate_koji_tag_bad_tag_value('forbiddentagprefix-foo')

    def test_validate_koji_tag_bad_tag_value_in_list(self):
        """ Test that we fail on a list containing bad tag value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_bad_tag_value_in_list(tag_arg):
            pass

        with pytest.raises(ValidationError):
            validate_koji_tag_bad_tag_value_in_list([
                'module-foo', 'forbiddentagprefix-bar'])

    def test_validate_koji_tag_good_tag_value(self):
        """ Test that we pass on a good tag value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_good_tag_value(tag_arg):
            return True

        assert validate_koji_tag_good_tag_value('module-foo') is True

    def test_validate_koji_tag_good_tag_values_in_list(self):
        """ Test that we pass on a list of good tag values. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_good_tag_values_in_list(tag_arg):
            return True

        assert validate_koji_tag_good_tag_values_in_list(['module-foo', 'module-bar']) is True

    def test_validate_koji_tag_good_tag_value_in_dict(self):
        """ Test that we pass on a dict arg with default key
        and a good value. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_good_tag_value_in_dict(tag_arg):
            return True

        assert validate_koji_tag_good_tag_value_in_dict({'name': 'module-foo'}) is True

    def test_validate_koji_tag_good_tag_value_in_dict_nondefault_key(self):
        """ Test that we pass on a dict arg with non-default key
        and a good value. """

        @module_build_service.utils.validate_koji_tag('tag_arg',
                                                      dict_key='nondefault')
        def validate_koji_tag_good_tag_value_in_dict_nondefault_key(tag_arg):
            return True

        assert validate_koji_tag_good_tag_value_in_dict_nondefault_key(
            {'nondefault': 'module-foo'}) is True

    def test_validate_koji_tag_double_trouble_good(self):
        """ Test that we pass on a list of tags that are good. """

        expected = 'foo'

        @module_build_service.utils.validate_koji_tag(['tag_arg1', 'tag_arg2'])
        def validate_koji_tag_double_trouble(tag_arg1, tag_arg2):
            return expected

        actual = validate_koji_tag_double_trouble('module-1', 'module-2')
        assert actual == expected

    def test_validate_koji_tag_double_trouble_bad(self):
        """ Test that we fail on a list of tags that are bad. """

        @module_build_service.utils.validate_koji_tag(['tag_arg1', 'tag_arg2'])
        def validate_koji_tag_double_trouble(tag_arg1, tag_arg2):
            pass

        with pytest.raises(ValidationError):
            validate_koji_tag_double_trouble('module-1', 'BADNEWS-2')

    def test_validate_koji_tag_is_None(self):
        """ Test that we fail on a tag which is None. """

        @module_build_service.utils.validate_koji_tag('tag_arg')
        def validate_koji_tag_is_None(tag_arg):
            pass

        with pytest.raises(ValidationError) as cm:
            validate_koji_tag_is_None(None)
            assert str(cm.value).endswith(' No value provided.') is True

    @patch('module_build_service.scm.SCM')
    def test_record_component_builds_duplicate_components(self, mocked_scm):
        with app.app_context():
            clean_database()
            mocked_scm.return_value.commit = \
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4'
            mocked_scm.return_value.get_latest.side_effect = [
                '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
                'fbed359411a1baa08d4a88e0d12d426fbf8f602c']

            testmodule_mmd_path = path.join(BASE_DIR, '..', 'staged_data', 'testmodule.yaml')
            mmd = Modulemd.Module().new_from_file(testmodule_mmd_path)
            mmd.upgrade()
            mmd.set_name('testmodule-variant')
            module_build = module_build_service.models.ModuleBuild()
            module_build.name = 'testmodule-variant'
            module_build.stream = 'master'
            module_build.version = 20170109091357
            module_build.state = models.BUILD_STATES['init']
            module_build.scmurl = 'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79'
            module_build.batch = 1
            module_build.owner = 'Tom Brady'
            module_build.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
            module_build.time_modified = datetime(2017, 2, 15, 16, 19, 35)
            module_build.rebuild_strategy = 'changed-and-after'
            module_build.modulemd = mmd.dumps()
            db.session.add(module_build)
            db.session.commit()
            # Rename the the modulemd to include
            mmd.set_name('testmodule')
            # Remove perl-Tangerine and tangerine from the modulemd to include so only one
            # component conflicts
            comps = mmd.get_rpm_components()
            del comps['perl-Tangerine']
            del comps['tangerine']
            mmd.set_rpm_components(comps)

            error_msg = (
                'The included module "testmodule" in "testmodule-variant" have '
                'the following conflicting components: perl-List-Compare')
            with pytest.raises(UnprocessableEntity) as e:
                module_build_service.utils.record_component_builds(
                    mmd, module_build, main_mmd=module_build.mmd())

            assert str(e.value) == error_msg

    @patch('module_build_service.scm.SCM')
    def test_record_component_builds_set_weight(self, mocked_scm):
        with app.app_context():
            clean_database()
            mocked_scm.return_value.commit = \
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4'
            mocked_scm.return_value.get_latest.side_effect = [
                '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
                'fbed359411a1baa08d4a88e0d12d426fbf8f602c',
                'dbed259411a1baa08d4a88e0d12d426fbf8f6037']

            testmodule_mmd_path = path.join(
                BASE_DIR, '..', 'staged_data', 'testmodule.yaml')
            mmd = Modulemd.Module().new_from_file(testmodule_mmd_path)
            mmd.upgrade()
            module_build = module_build_service.models.ModuleBuild()
            module_build.name = 'testmodule'
            module_build.stream = 'master'
            module_build.version = 20170109091357
            module_build.state = models.BUILD_STATES['init']
            module_build.scmurl = \
                'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79'
            module_build.batch = 1
            module_build.owner = 'Tom Brady'
            module_build.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
            module_build.time_modified = datetime(2017, 2, 15, 16, 19, 35)
            module_build.rebuild_strategy = 'changed-and-after'
            module_build.modulemd = mmd.dumps()
            db.session.add(module_build)
            db.session.commit()

            module_build_service.utils.record_component_builds(mmd, module_build)

            assert module_build.state == models.BUILD_STATES['init']
            db.session.refresh(module_build)
            for c in module_build.component_builds:
                assert c.weight == 1.5

    @patch('module_build_service.scm.SCM')
    def test_format_mmd_arches(self, mocked_scm):
        with app.app_context():
            clean_database()
            mocked_scm.return_value.commit = \
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4'
            mocked_scm.return_value.get_latest.side_effect = [
                '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
                'fbed359411a1baa08d4a88e0d12d426fbf8f602c',
                'dbed259411a1baa08d4a88e0d12d426fbf8f6037',
                '4ceea43add2366d8b8c5a622a2fb563b625b9abf',
                'fbed359411a1baa08d4a88e0d12d426fbf8f602c',
                'dbed259411a1baa08d4a88e0d12d426fbf8f6037']

            testmodule_mmd_path = path.join(
                BASE_DIR, '..', 'staged_data', 'testmodule.yaml')
            test_archs = ['powerpc', 'i486']

            mmd1 = Modulemd.Module().new_from_file(testmodule_mmd_path)
            mmd1.upgrade()

            module_build_service.utils.format_mmd(mmd1, None)

            for pkg in mmd1.get_rpm_components().values():
                assert set(pkg.get_arches().get()) == set(conf.arches)

            mmd2 = Modulemd.Module().new_from_file(testmodule_mmd_path)
            mmd2.upgrade()

            new_arches = Modulemd.SimpleSet()
            new_arches.set(test_archs)
            for pkg in mmd2.get_rpm_components().values():
                pkg.set_arches(new_arches)

            module_build_service.utils.format_mmd(mmd2, None)

            for pkg in mmd2.get_rpm_components().values():
                assert set(pkg.get_arches().get()) == set(test_archs)

    def test_generate_koji_tag_in_nsvc_format(self):
        name, stream, version, context = ('testmodule', 'master', '20170816080815', '37c6c57')

        tag = module_build_service.utils.generate_koji_tag(name, stream, version, context)

        assert tag == 'module-testmodule-master-20170816080815-37c6c57'

    def test_generate_koji_tag_in_hash_format(self):
        name, version, context = ('testmodule', '20170816080815', '37c6c57')
        stream = 'this-is-a-stream-with-very-looooong-name' + '-blah' * 50
        nsvc_list = [name, stream, version, context]

        tag = module_build_service.utils.generate_koji_tag(*nsvc_list)
        expected_tag = 'module-1cf457d452e54dda'
        assert tag == expected_tag

    @patch("module_build_service.utils.submit.requests")
    def test_pdc_eol_check(self, requests):
        """ Push mock pdc responses through the eol check function. """

        response = mock.Mock()
        response.json.return_value = {"results": [{
            "id": 347907,
            "global_component": "mariadb",
            "name": "10.1",
            "slas": [{
                "id": 694207,
                "sla": "security_fixes",
                "eol": "2019-12-01",
            }],
            "type": "module",
            "active": True,
            "critical_path": False,
        }]}
        requests.get.return_value = response

        is_eol = module_build_service.utils.submit._is_eol_in_pdc('mariadb', '10.1')
        assert not is_eol

        response.json.return_value["results"][0]["active"] = False

        is_eol = module_build_service.utils.submit._is_eol_in_pdc('mariadb', '10.1')
        assert is_eol

    def test_get_prefixed_version_f28(self):
        scheduler_init_data(1)
        build_one = models.ModuleBuild.query.get(2)
        v = module_build_service.utils.submit.get_prefixed_version(build_one.mmd())
        assert v == 2820180205135154

    def test_get_prefixed_version_fl701(self):
        scheduler_init_data(1)
        build_one = models.ModuleBuild.query.get(2)
        mmd = build_one.mmd()
        xmd = glib.from_variant_dict(mmd.get_xmd())
        xmd['mbs']['buildrequires']['platform']['stream'] = 'fl7.0.1-beta'
        mmd.set_xmd(glib.dict_values(xmd))
        v = module_build_service.utils.submit.get_prefixed_version(mmd)
        assert v == 7000120180205135154


class DummyModuleBuilder(GenericBuilder):
    """
    Dummy module builder
    """

    backend = "koji"
    _build_id = 0

    TAGGED_COMPONENTS = []

    @module_build_service.utils.validate_koji_tag('tag_name')
    def __init__(self, owner, module, config, tag_name, components):
        self.module_str = module
        self.tag_name = tag_name
        self.config = config

    def buildroot_connect(self, groups):
        pass

    def buildroot_prep(self):
        pass

    def buildroot_resume(self):
        pass

    def buildroot_ready(self, artifacts=None):
        return True

    def buildroot_add_dependency(self, dependencies):
        pass

    def buildroot_add_artifacts(self, artifacts, install=False):
        DummyModuleBuilder.TAGGED_COMPONENTS += artifacts

    def buildroot_add_repos(self, dependencies):
        pass

    def tag_artifacts(self, artifacts):
        pass

    def recover_orphaned_artifact(self, component_build):
        return []

    @property
    def module_build_tag(self):
        return {"name": self.tag_name + "-build"}

    def build(self, artifact_name, source):
        DummyModuleBuilder._build_id += 1
        state = koji.BUILD_STATES['COMPLETE']
        reason = "Submitted %s to Koji" % (artifact_name)
        return DummyModuleBuilder._build_id, state, reason, None

    @staticmethod
    def get_disttag_srpm(disttag, module_build):
        # @FIXME
        return KojiModuleBuilder.get_disttag_srpm(disttag, module_build)

    def cancel_build(self, task_id):
        pass

    def list_tasks_for_components(self, component_builds=None, state='active'):
        pass

    def repo_from_tag(self, config, tag_name, arch):
        pass

    def finalize(self):
        pass


@patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
       return_value={'build': [], 'srpm-build': []})
class TestBatches:

    def setup_method(self, test_method):
        reuse_component_init_data()
        GenericBuilder.register_backend_class(DummyModuleBuilder)

    def teardown_method(self, test_method):
        clean_database()
        DummyModuleBuilder.TAGGED_COMPONENTS = []
        GenericBuilder.register_backend_class(KojiModuleBuilder)

    def test_start_next_batch_build_reuse(self, default_buildroot_groups):
        """
        Tests that start_next_batch_build:
           1) Increments module.batch.
           2) Can reuse all components in batch
           3) Returns proper further_work messages for reused components.
           4) Returns the fake Repo change message
           5) Handling the further_work messages lead to proper tagging of
              reused components.
        """
        module_build = models.ModuleBuild.query.filter_by(id=3).one()
        module_build.batch = 1

        builder = mock.MagicMock()
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        assert module_build.batch == 2

        # KojiBuildChange messages in further_work should have build_new_state
        # set to COMPLETE, but the current component build state should be set
        # to BUILDING, so KojiBuildChange message handler handles the change
        # properly.
        for msg in further_work:
            if type(msg) == module_build_service.messaging.KojiBuildChange:
                assert msg.build_new_state == koji.BUILD_STATES['COMPLETE']
                component_build = models.ComponentBuild.from_component_event(db.session, msg)
                assert component_build.state == koji.BUILD_STATES['BUILDING']

        # When we handle these KojiBuildChange messages, MBS should tag all
        # the components just once.
        for msg in further_work:
            if type(msg) == module_build_service.messaging.KojiBuildChange:
                module_build_service.scheduler.handlers.components.complete(
                    conf, db.session, msg)

        # Since we have reused all the components in the batch, there should
        # be fake KojiRepoChange message.
        assert type(further_work[-1]) == module_build_service.messaging.KojiRepoChange

        # Check that packages have been tagged just once.
        assert len(DummyModuleBuilder.TAGGED_COMPONENTS) == 2

    @patch('module_build_service.utils.batches.start_build_component')
    def test_start_next_batch_build_reuse_some(self, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build:
           1) Increments module.batch.
           2) Can reuse all components in the batch that it can.
           3) Returns proper further_work messages for reused components.
           4) Builds the remaining components
           5) Handling the further_work messages lead to proper tagging of
              reused components.
        """
        module_build = models.ModuleBuild.query.filter_by(id=3).one()
        module_build.batch = 1
        plc_component = models.ComponentBuild.query.filter_by(
            module_id=3, package='perl-List-Compare').one()
        plc_component.ref = '5ceea46add2366d8b8c5a623a2fb563b625b9abd'

        builder = mock.MagicMock()
        builder.recover_orphaned_artifact.return_value = []
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        assert module_build.batch == 2

        # Make sure we only have one message returned for the one reused component
        assert len(further_work) == 1
        # The KojiBuildChange message in further_work should have build_new_state
        # set to COMPLETE, but the current component build state in the DB should be set
        # to BUILDING, so KojiBuildChange message handler handles the change
        # properly.
        assert further_work[0].build_new_state == koji.BUILD_STATES['COMPLETE']
        component_build = models.ComponentBuild.from_component_event(db.session, further_work[0])
        assert component_build.state == koji.BUILD_STATES['BUILDING']
        assert component_build.package == 'perl-Tangerine'
        assert component_build.reused_component_id is not None
        # Make sure perl-List-Compare is set to the build state as well but not reused
        assert plc_component.state == koji.BUILD_STATES['BUILDING']
        assert plc_component.reused_component_id is None
        mock_sbc.assert_called_once()

    @patch('module_build_service.utils.batches.start_build_component')
    @patch('module_build_service.config.Config.rebuild_strategy',
           new_callable=mock.PropertyMock, return_value='all')
    def test_start_next_batch_build_rebuild_strategy_all(
            self, mock_rm, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build can't reuse any components in the batch because the
        rebuild method is set to "all".
        """
        module_build = models.ModuleBuild.query.filter_by(id=3).one()
        module_build.rebuild_strategy = 'all'
        module_build.batch = 1

        builder = mock.MagicMock()
        builder.recover_orphaned_artifact.return_value = []
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        assert module_build.batch == 2
        # No component reuse messages should be returned
        assert len(further_work) == 0
        # Make sure that both components in the batch were submitted
        assert len(mock_sbc.mock_calls) == 2

    def test_start_build_component_failed_state(self, default_buildroot_groups):
        """
        Tests whether exception occured while building sets the state to failed
        """
        builder = mock.MagicMock()
        builder.build.side_effect = Exception('Something have gone terribly wrong')
        component = mock.MagicMock()

        module_build_service.utils.batches.start_build_component(builder, component)

        assert component.state == koji.BUILD_STATES['FAILED']

    @patch('module_build_service.utils.batches.start_build_component')
    @patch('module_build_service.config.Config.rebuild_strategy',
           new_callable=mock.PropertyMock, return_value='only-changed')
    def test_start_next_batch_build_rebuild_strategy_only_changed(
            self, mock_rm, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build reuses all unchanged components in the batch because the
        rebuild method is set to "only-changed". This means that one component is reused in batch
        2, and even though the other component in batch 2 changed and was rebuilt, the component
        in batch 3 can be reused.
        """
        module_build = models.ModuleBuild.query.filter_by(id=3).one()
        module_build.rebuild_strategy = 'only-changed'
        module_build.batch = 1
        # perl-List-Compare changed
        plc_component = models.ComponentBuild.query.filter_by(
            module_id=3, package='perl-List-Compare').one()
        plc_component.ref = '5ceea46add2366d8b8c5a623a2fb563b625b9abd'

        builder = mock.MagicMock()
        builder.recover_orphaned_artifact.return_value = []
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase
        assert module_build.batch == 2

        # Make sure we only have one message returned for the one reused component
        assert len(further_work) == 1
        # The KojiBuildChange message in further_work should have build_new_state
        # set to COMPLETE, but the current component build state in the DB should be set
        # to BUILDING, so KojiBuildChange message handler handles the change
        # properly.
        assert further_work[0].build_new_state == koji.BUILD_STATES['COMPLETE']
        component_build = models.ComponentBuild.from_component_event(db.session, further_work[0])
        assert component_build.state == koji.BUILD_STATES['BUILDING']
        assert component_build.package == 'perl-Tangerine'
        assert component_build.reused_component_id is not None
        # Make sure perl-List-Compare is set to the build state as well but not reused
        assert plc_component.state == koji.BUILD_STATES['BUILDING']
        assert plc_component.reused_component_id is None
        mock_sbc.assert_called_once()
        mock_sbc.reset_mock()

        # Complete the build
        plc_component.state = koji.BUILD_STATES['COMPLETE']
        pt_component = models.ComponentBuild.query.filter_by(
            module_id=3, package='perl-Tangerine').one()
        pt_component.state = koji.BUILD_STATES['COMPLETE']

        # Start the next build batch
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)
        # Batch number should increase
        assert module_build.batch == 3
        # Verify that tangerine was reused even though perl-Tangerine was rebuilt in the previous
        # batch
        assert further_work[0].build_new_state == koji.BUILD_STATES['COMPLETE']
        component_build = models.ComponentBuild.from_component_event(db.session, further_work[0])
        assert component_build.state == koji.BUILD_STATES['BUILDING']
        assert component_build.package == 'tangerine'
        assert component_build.reused_component_id is not None
        mock_sbc.assert_not_called()

    @patch('module_build_service.utils.batches.start_build_component')
    def test_start_next_batch_build_smart_scheduling(self, mock_sbc, default_buildroot_groups):
        """
        Tests that components with the longest build time will be scheduled first
        """
        module_build = models.ModuleBuild.query.filter_by(id=3).one()
        module_build.batch = 1
        pt_component = models.ComponentBuild.query.filter_by(
            module_id=3, package='perl-Tangerine').one()
        pt_component.ref = '6ceea46add2366d8b8c5a623b2fb563b625bfabe'
        plc_component = models.ComponentBuild.query.filter_by(
            module_id=3, package='perl-List-Compare').one()
        plc_component.ref = '5ceea46add2366d8b8c5a623a2fb563b625b9abd'

        # Components are by default built by component id. To find out that weight is respected,
        # we have to set bigger weight to component with lower id.
        pt_component.weight = 3 if pt_component.id < plc_component.id else 4
        plc_component.weight = 4 if pt_component.id < plc_component.id else 3

        builder = mock.MagicMock()
        builder.recover_orphaned_artifact.return_value = []
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should increase.
        assert module_build.batch == 2

        # Make sure we don't have any messages returned since no components should be reused
        assert len(further_work) == 0
        # Make sure both components are set to the build state but not reused
        assert pt_component.state == koji.BUILD_STATES['BUILDING']
        assert pt_component.reused_component_id is None
        assert plc_component.state == koji.BUILD_STATES['BUILDING']
        assert plc_component.reused_component_id is None

        # Test the order of the scheduling
        expected_calls = [mock.call(builder, plc_component), mock.call(builder, pt_component)]
        assert mock_sbc.mock_calls == expected_calls

    @patch('module_build_service.utils.batches.start_build_component')
    def test_start_next_batch_continue(self, mock_sbc, default_buildroot_groups):
        """
        Tests that start_next_batch_build does not start new batch when
        there are unbuilt components in the current one.
        """
        module_build = models.ModuleBuild.query.filter_by(id=3).one()
        module_build.batch = 2

        # The component was reused when the batch first started
        building_component = module_build.current_batch()[0]
        building_component.state = koji.BUILD_STATES['BUILDING']
        building_component.reused_component_id = 123
        db.session.commit()

        builder = mock.MagicMock()
        further_work = module_build_service.utils.start_next_batch_build(
            conf, module_build, db.session, builder)

        # Batch number should not increase.
        assert module_build.batch == 2
        # Make sure start build was called for the second component which wasn't reused
        mock_sbc.assert_called_once()
        # No further work should be returned
        assert len(further_work) == 0

    def test_start_next_batch_build_repo_building(self, default_buildroot_groups):
        """
        Test that start_next_batch_build does not start new batch when
        builder.buildroot_ready() returns False.
        """
        module_build = models.ModuleBuild.query.filter_by(id=3).one()
        module_build.batch = 1

        builder = mock.MagicMock()
        builder.buildroot_ready.return_value = False

        # Batch number should not increase.
        assert module_build.batch == 1


@patch("module_build_service.config.Config.mock_resultsdir",
       new_callable=mock.PropertyMock,
       return_value=path.join(
           BASE_DIR, '..', 'staged_data', "local_builds"))
@patch("module_build_service.config.Config.system",
       new_callable=mock.PropertyMock, return_value="mock")
class TestLocalBuilds:

    def setup_method(self):
        clean_database()

    def teardown_method(self):
        clean_database()

    def test_load_local_builds_name(self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("testmodule")
            local_modules = models.ModuleBuild.local_modules(db.session)

            assert len(local_modules) == 1
            assert local_modules[0].koji_tag.endswith(
                "/module-testmodule-master-20170816080816/results")

    def test_load_local_builds_name_stream(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("testmodule:master")
            local_modules = models.ModuleBuild.local_modules(db.session)

            assert len(local_modules) == 1
            assert local_modules[0].koji_tag.endswith(
                "/module-testmodule-master-20170816080816/results")

    def test_load_local_builds_name_stream_non_existing(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            with pytest.raises(RuntimeError):
                module_build_service.utils.load_local_builds("testmodule:x")
                models.ModuleBuild.local_modules(db.session)

    def test_load_local_builds_name_stream_version(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("testmodule:master:20170816080815")
            local_modules = models.ModuleBuild.local_modules(db.session)

            assert len(local_modules) == 1
            assert local_modules[0].koji_tag.endswith(
                "/module-testmodule-master-20170816080815/results")

    def test_load_local_builds_name_stream_version_non_existing(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            with pytest.raises(RuntimeError):
                module_build_service.utils.load_local_builds("testmodule:master:123")
                models.ModuleBuild.local_modules(db.session)

    def test_load_local_builds_platform(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("platform")
            local_modules = models.ModuleBuild.local_modules(db.session)

            assert len(local_modules) == 1
            assert local_modules[0].koji_tag.endswith(
                "/module-platform-f28-3/results")

    def test_load_local_builds_platform_f28(
            self, conf_system, conf_resultsdir):
        with app.app_context():
            module_build_service.utils.load_local_builds("platform:f28")
            local_modules = models.ModuleBuild.local_modules(db.session)

            assert len(local_modules) == 1
            assert local_modules[0].koji_tag.endswith(
                "/module-platform-f28-3/results")



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Petr Šabata <contyk@redhat.com>
#            Luboš Kocman <lkocman@redhat.com>

"""Generic component build functions."""

# TODO: Query the MBS to find what modules satisfy the build dependencies and
#       their tag names.
# TODO: Ensure the RPM %dist tag is set according to the policy.

import six
import dogpile.cache
from abc import ABCMeta, abstractmethod
from requests.exceptions import ConnectionError

from module_build_service import conf, log, db
from module_build_service import models
import module_build_service.resolver
import module_build_service.scm
import module_build_service.utils
from module_build_service.utils import create_dogpile_key_generator_func


"""
Example workflows - helps to see the difference in implementations
Koji workflow

1) create tag, and build-tag
2) create target out of ^tag and ^build-tag
3) run regen-repo to have initial repodata (happens automatically)
4) build module-build-macros which provides "dist" macro
5) tag module-build-macro into buildroot
6) wait for module-build-macro to be available in buildroot
7) build all components from scmurl
8) (optional) wait for selected builds to be available in buildroot

"""


class GenericBuilder(six.with_metaclass(ABCMeta)):
    """
    External Api for builders

    Example usage:
        config = module_build_service.config.Config()
        builder = Builder(module="testmodule-1.2-3", backend="koji", config)
        builder.buildroot_connect()
        builder.build(artifact_name="bash",
                      source="git://pkgs.stg.fedoraproject.org/rpms/bash"
                             "?#70fa7516b83768595a4f3280ae890a7ac957e0c7")

        ...
        # E.g. on some other worker ... just resume buildroot that was initially created
        builder = Builder(module="testmodule-1.2-3", backend="koji", config)
        builder.buildroot_connect()
        builder.build(artifact_name="not-bash",
                      source="git://pkgs.stg.fedoraproject.org/rpms/not-bash"
                             "?#70fa7516b83768595a4f3280ae890a7ac957e0c7")
        # wait until this particular bash is available in the buildroot
        builder.buildroot_ready(artifacts=["bash-1.23-el6"])
        builder.build(artifact_name="not-not-bash",
                      source="git://pkgs.stg.fedoraproject.org/rpms/not-not-bash"
                             "?#70fa7516b83768595a4f3280ae890a7ac957e0c7")

    """

    backend = "generic"
    backends = {}

    # Create region to cache the default_buildroot_groups results.
    # We are skipping the caching based on the first two arguments of
    # default_buildroot_groups, because they are "self" and db.session
    # instance which are different each call we call that method.
    default_buildroot_groups_cache = dogpile.cache.make_region(
        function_key_generator=create_dogpile_key_generator_func(2)).configure(
            'dogpile.cache.memory')

    @classmethod
    def register_backend_class(cls, backend_class):
        GenericBuilder.backends[backend_class.backend] = backend_class

    @classmethod
    def create(cls, owner, module, backend, config, **extra):
        """
        :param owner: a string representing who kicked off the builds
        :param module: module_build_service.models.ModuleBuild instance.
        :param backend: a string representing backend e.g. 'koji'
        :param config: instance of module_build_service.config.Config

        Any additional arguments are optional extras which can be passed along
        and are implementation-dependent.
        """
        # check if the backend is within allowed backends for the used resolver
        resolver = module_build_service.resolver.system_resolver
        if not resolver.is_builder_compatible(backend):
            raise ValueError("Builder backend '{}' is not compatible with "
                             "resolver backend '{}'. Check your configuration."
                             .format(backend, resolver.backend))

        if backend in GenericBuilder.backends:
            return GenericBuilder.backends[backend](owner=owner, module=module,
                                                    config=config, **extra)
        else:
            raise ValueError("Builder backend='%s' not recognized" % backend)

    @classmethod
    def create_from_module(cls, session, module, config, buildroot_connect=True):
        """
        Creates new GenericBuilder instance based on the data from module
        and config and connects it to buildroot.

        :param session: SQLAlchemy databa session.
        :param module: module_build_service.models.ModuleBuild instance.
        :param config: module_build_service.config.Config instance.
        :kwarg buildroot_connect: a boolean that determines if the builder should run
        buildroot_connect on instantiation.
        """
        components = [c.package for c in module.component_builds]
        builder = GenericBuilder.create(
            module.owner, module, config.system, config, tag_name=module.koji_tag,
            components=components)
        if buildroot_connect is True:
            groups = GenericBuilder.default_buildroot_groups(session, module)
            builder.buildroot_connect(groups)
        return builder

    @classmethod
    def tag_to_repo(cls, backend, config, tag_name, arch):
        """
        :param backend: a string representing the backend e.g. 'koji'.
        :param config: instance of module_build_service.config.Config
        :param tag_name: Tag for which the repository is returned
        :param arch: Architecture for which the repository is returned

        Returns URL of repository containing the built artifacts for
        the tag with particular name and architecture.
        """
        if backend in GenericBuilder.backends:
            return GenericBuilder.backends[backend].repo_from_tag(
                config, tag_name, arch)
        else:
            raise ValueError("Builder backend='%s' not recognized" % backend)

    @abstractmethod
    def buildroot_connect(self, groups):
        """
        This is an idempotent call to create or resume and validate the build
        environment.  .build() should immediately fail if .buildroot_connect()
        wasn't called.

        Koji Example: create tag, targets, set build tag inheritance...
        """
        raise NotImplementedError()

    @abstractmethod
    def buildroot_ready(self, artifacts=None):
        """
        :param artifacts=None : a list of artifacts supposed to be in the buildroot
                                (['bash-123-0.el6'])

        returns when the buildroot is ready (or contains the specified artifact)

        This function is here to ensure that the buildroot (repo) is ready and
        contains the listed artifacts if specified.
        """
        raise NotImplementedError()

    @abstractmethod
    def buildroot_add_repos(self, dependencies):
        """
        :param dependencies: a list of modules represented as a list of dicts,
                             like:
                             [{'name': ..., 'version': ..., 'release': ...}, ...]

        Make an additional repository available in the buildroot. This does not
        necessarily have to directly install artifacts (e.g. koji), just make
        them available.

        E.g. the koji implementation of the call uses MBS to get koji_tag
        associated with each module dep and adds the tag to $module-build tag
        inheritance.
        """
        raise NotImplementedError()

    @abstractmethod
    def buildroot_add_artifacts(self, artifacts, install=False):
        """
        :param artifacts: list of artifacts to be available or installed
                          (install=False) in the buildroot (e.g  list of $NEVRAS)
        :param install=False: pre-install artifact in the buildroot (otherwise
                              "just make it available for install")

        Example:

        koji tag-build $module-build-tag bash-1.234-1.el6
        if install:
            koji add-group-pkg $module-build-tag build bash
            # This forces install of bash into buildroot and srpm-buildroot
            koji add-group-pkg $module-build-tag srpm-build bash
        """
        raise NotImplementedError()

    @abstractmethod
    def tag_artifacts(self, artifacts):
        """
        :param artifacts: list of artifacts (NVRs) to be tagged

        Adds the artifacts to tag associated with this module build.
        """
        raise NotImplementedError()

    @abstractmethod
    def build(self, artifact_name, source):
        """
        :param artifact_name : A package name. We can't guess it since macros
                               in the buildroot could affect it, (e.g. software
                               collections).
        :param source : an SCM URL, clearly identifying the build artifact in a
                        repository
        :return 4-tuple of the form (build task id, state, reason, nvr)

        The artifact_name parameter is used in koji add-pkg (and it's actually
        the only reason why we need to pass it). We don't really limit source
        types. The actual source is usually delivered as an SCM URL from
        fedmsg.

        Warning: This function must be thread-safe.

        Example
        .build("bash", "git://someurl/bash#damn") #build from SCM URL
        .build("bash", "/path/to/srpm.src.rpm") #build from source RPM
        """
        raise NotImplementedError()

    @abstractmethod
    def cancel_build(self, task_id):
        """
        :param task_id: Task ID returned by the build method.

        Cancels the build.
        """
        raise NotImplementedError()

    @abstractmethod
    def finalize(self):
        """
        :return: None

        This method is supposed to be called after all module builds are
        successfully finished.

        It could be utilized for various purposes such as cleaning or
        running additional build-system based operations on top of
        finished builds
        """
        pass

    @classmethod
    @abstractmethod
    def repo_from_tag(self, config, tag_name, arch):
        """
        :param config: instance of module_build_service.config.Config
        :param tag_name: Tag for which the repository is returned
        :param arch: Architecture for which the repository is returned

        Returns URL of repository containing the built artifacts for
        the tag with particular name and architecture.
        """
        raise NotImplementedError()

    @classmethod
    def clear_cache(cls, module_build):
        """
        Clears the per module build default_buildroot_groups cache.
        """
        cls.default_buildroot_groups_cache.delete(
            "default_buildroot_groups_" + str(module_build.id))

    @classmethod
    @module_build_service.utils.retry(wait_on=(ConnectionError))
    @default_buildroot_groups_cache.cache_on_arguments()
    def default_buildroot_groups(cls, session, module):
        local_modules = models.ModuleBuild.local_modules(db.session)
        local_modules = {m.name + "-" + m.stream: m for m in local_modules}

        try:
            mmd = module.mmd()
            resolver = module_build_service.resolver.system_resolver

            # Resolve default buildroot groups using the MBS, but only for
            # non-local modules.
            groups = resolver.resolve_profiles(
                mmd, ('buildroot', 'srpm-buildroot'))

            groups = {
                'build': groups['buildroot'],
                'srpm-build': groups['srpm-buildroot'],
            }
        except ValueError:
            reason = "Failed to gather buildroot groups from SCM."
            log.exception(reason)
            module.transition(conf, state="failed", state_reason=reason)
            session.commit()
            raise
        return groups

    @abstractmethod
    def list_tasks_for_components(self, component_builds=None, state='active'):
        """
        :param component_builds: list of component builds which we want to check
        :param state: limit the check only for tasks in the given state
        :return: list of tasks

        This method is supposed to list tasks ('active' by default)
        for component builds.
        """
        raise NotImplementedError()

    @classmethod
    def get_built_rpms_in_module_build(cls, mmd):
        """
        :param Modulemd mmd: Modulemd to get the built RPMs from.
        :return: list of NVRs
        """
        raise NotImplementedError()

    @classmethod
    def recover_orphaned_artifact(cls, component_build):
        """
        Searches for a complete build of an artifact belonging to the module and sets the
        component_build in the MBS database to the found build. This usually returns nothing since
        these builds should *not* exist.
        :param artifact_name: a ComponentBuild object
        :return: a list of msgs that MBS needs to process
        """
        return []

    @classmethod
    def get_average_build_time(self, component):
        """
        Placeholder function for the builders to report the average time it takes to build the
        specified component. If this function is not overridden, then 0.0 is returned.
        :param component: a ComponentBuild object
        :return: a float of 0.0
        """
        return 0.0

    @classmethod
    def get_build_weights(cls, components):
        """
        Returns a dict with component name as a key and float number
        representing the overall Koji weight of a component build.

        :param list components: List of component names.
        :rtype: dict
        :return: {component_name: weight_as_float, ...}
        """
        return cls.compute_weights_from_build_time(components)

    @classmethod
    def compute_weights_from_build_time(cls, components, arches=None):
        """
        Computes the weights of ComponentBuilds based on average time to build
        and list of arches for which the component is going to be built.

        This method should be used as a fallback only when KojiModuleBuilder
        cannot be used, because the weight this method produces is not 100% accurate.

        :param components: List of comopnent names to compute the weight for.
        :param arches: List of arches to build for or None. If the value is None,
            conf.arches will be used instead.
        :rtype: dict
        :return: {component_name: weight_as_float, ...}
        """
        if not arches:
            arches = conf.arches

        weights = {}

        for component in components:
            average_time_to_build = cls.get_average_build_time(component)

            # The way how `weight` is computed is based on hardcoded weight values
            # in kojid.py.
            # The weight computed here is not 100% accurate, because there are
            # multiple smaller tasks in koji like waitrepo or createrepo and we
            # cannot say if they will be executed as part of this component build.
            # The weight computed here is used only to limit the number of builds
            # and we generally do not care about waitrepo/createrepo weights in MBS.

            # 1.5 is what Koji hardcodes as a default weight for BuildArchTask.
            weight = 1.5
            if not average_time_to_build:
                weights[component] = weight
                continue

            if average_time_to_build < 0:
                log.warning("Negative average build duration for component %s: %s",
                            component, str(average_time_to_build))
                weights[component] = weight
                continue

            # Increase the task weight by 0.75 for every hour of build duration.
            adj = (average_time_to_build / ((60 * 60) / 0.75))
            # cap the adjustment at +4.5
            weight += min(4.5, adj)

            # We build for all arches, so multiply the weight by number of arches.
            weight = weight * len(arches)

            # 1.5 here is hardcoded Koji weight of single BuildSRPMFromSCMTask
            weight += 1.5
            weights[component] = weight

        return weights



# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

from mock import patch, PropertyMock, Mock, call

import module_build_service.resolver as mbs_resolver
import module_build_service.utils
import module_build_service.models
from module_build_service import glib, Modulemd, app
import tests


base_dir = os.path.join(os.path.dirname(__file__), "..")


class TestMBSModule:

    @patch("requests.Session")
    def test_get_module_modulemds_nsvc(self, mock_session, testmodule_mmd_9c690d0e):
        """ Tests for querying a module from mbs """
        mock_res = Mock()
        mock_res.ok.return_value = True
        mock_res.json.return_value = {
            "items": [
                {
                    "name": "testmodule",
                    "stream": "master",
                    "version": "20180205135154",
                    "context": "9c690d0e",
                    "modulemd": testmodule_mmd_9c690d0e
                }
            ],
            "meta": {"next": None}
        }

        mock_session().get.return_value = mock_res

        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='mbs')
        module_mmds = resolver.get_module_modulemds('testmodule', 'master', '20180205135154',
                                                    '9c690d0e')
        nsvcs = set(
            '{}:{}:{}:{}'.format(m.peek_name(), m.peek_stream(),
                                 m.peek_version(), m.peek_context())
            for m in module_mmds
        )
        expected = set(["testmodule:master:20180205135154:9c690d0e"])
        mbs_url = tests.conf.mbs_url
        expected_query = {
            "name": "testmodule",
            "stream": "master",
            "version": "20180205135154",
            "context": "9c690d0e",
            "verbose": True,
            "order_desc_by": "version",
            "page": 1,
            "per_page": 10,
            "state": "ready"
        }
        mock_session().get.assert_called_once_with(mbs_url, params=expected_query)
        assert nsvcs == expected

    @patch("requests.Session")
    def test_get_module_modulemds_partial(self, mock_session, testmodule_mmd_9c690d0e,
                                          testmodule_mmd_c2c572ed):
        """ Test for querying MBS without the context of a module """

        version = "20180205135154"

        mock_res = Mock()
        mock_res.ok.return_value = True
        mock_res.json.return_value = {
            "items": [
                {
                    "name": "testmodule",
                    "stream": "master",
                    "version": version,
                    "context": "9c690d0e",
                    "modulemd": testmodule_mmd_9c690d0e
                },
                {
                    "name": "testmodule",
                    "stream": "master",
                    "version": version,
                    "context": "c2c572ed",
                    "modulemd": testmodule_mmd_c2c572ed
                }
            ],
            "meta": {"next": None}
        }

        mock_session().get.return_value = mock_res
        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='mbs')
        ret = resolver.get_module_modulemds('testmodule', 'master', version)
        nsvcs = set(
            '{}:{}:{}:{}'.format(m.peek_name(), m.peek_stream(),
                                 m.peek_version(), m.peek_context())
            for m in ret
        )
        expected = set(["testmodule:master:20180205135154:9c690d0e",
                        "testmodule:master:20180205135154:c2c572ed"])
        mbs_url = tests.conf.mbs_url
        expected_query = {
            "name": "testmodule",
            "stream": "master",
            "version": version,
            "verbose": True,
            "order_desc_by": "version",
            "page": 1,
            "per_page": 10,
            "state": "ready"
        }
        mock_session().get.assert_called_once_with(mbs_url, params=expected_query)
        assert nsvcs == expected

    @patch("requests.Session")
    def test_get_module_build_dependencies(self, mock_session, platform_mmd,
                                           testmodule_mmd_9c690d0e):
        """
        Tests that we return just direct build-time dependencies of testmodule.
        """
        mock_res = Mock()
        mock_res.ok.return_value = True
        mock_res.json.side_effect = [
            {
                "items": [
                    {
                        "name": "testmodule",
                        "stream": "master",
                        "version": "20180205135154",
                        "context": "9c690d0e",
                        "modulemd": testmodule_mmd_9c690d0e
                    }
                ],
                "meta": {"next": None}
            }, {
                "items": [
                    {
                        "name": "platform",
                        "stream": "f28",
                        "version": "3",
                        "context": "00000000",
                        "modulemd": platform_mmd,
                        "koji_tag": "module-f28-build"
                    }
                ],
                "meta": {"next": None}
            }
        ]

        mock_session().get.return_value = mock_res
        expected = set(['module-f28-build'])
        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='mbs')
        result = resolver.get_module_build_dependencies(
            'testmodule', 'master', '20180205135154', '9c690d0e').keys()

        expected_queries = [{
            "name": "testmodule",
            "stream": "master",
            "version": "20180205135154",
            "context": "9c690d0e",
            "verbose": True,
            "order_desc_by": "version",
            "page": 1,
            "per_page": 10,
            "state": "ready"
        }, {
            "name": "platform",
            "stream": "f28",
            "version": "3",
            "context": "00000000",
            "verbose": True,
            "order_desc_by": "version",
            "page": 1,
            "per_page": 10,
            "state": "ready"
        }]

        mbs_url = tests.conf.mbs_url
        expected_calls = [call(mbs_url, params=expected_queries[0]),
                          call(mbs_url, params=expected_queries[1])]
        mock_session().get.mock_calls = expected_calls
        assert mock_session().get.call_count == 2
        assert set(result) == expected

    @patch("requests.Session")
    def test_get_module_build_dependencies_empty_buildrequires(self, mock_session,
                                                               testmodule_mmd_9c690d0e):

        mmd = Modulemd.Module().new_from_string(testmodule_mmd_9c690d0e)
        # Wipe out the dependencies
        mmd.set_dependencies()
        xmd = glib.from_variant_dict(mmd.get_xmd())
        xmd['mbs']['buildrequires'] = {}
        mmd.set_xmd(glib.dict_values(xmd))

        mock_res = Mock()
        mock_res.ok.return_value = True
        mock_res.json.side_effect = [
            {
                "items": [
                    {
                        "name": "testmodule",
                        "stream": "master",
                        "version": "20180205135154",
                        "context": "9c690d0e",
                        "modulemd": mmd.dumps(),
                        "build_deps": []
                    }
                ],
                "meta": {"next": None}
            }
        ]

        mock_session().get.return_value = mock_res

        expected = set()

        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='mbs')
        result = resolver.get_module_build_dependencies(
            'testmodule', 'master', '20180205135154', '9c690d0e').keys()
        mbs_url = tests.conf.mbs_url
        expected_query = {
            "name": "testmodule",
            "stream": "master",
            "version": "20180205135154",
            "context": "9c690d0e",
            "verbose": True,
            "order_desc_by": "version",
            "page": 1,
            "per_page": 10,
            "state": "ready"
        }
        mock_session().get.assert_called_once_with(mbs_url, params=expected_query)
        assert set(result) == expected

    @patch("requests.Session")
    def test_resolve_profiles(self, mock_session, formatted_testmodule_mmd, platform_mmd):

        mock_res = Mock()
        mock_res.ok.return_value = True
        mock_res.json.return_value = {
            "items": [
                {
                    "name": "platform",
                    "stream": "f28",
                    "version": "3",
                    "context": "00000000",
                    "modulemd": platform_mmd
                }
            ],
            "meta": {"next": None}
        }

        mock_session().get.return_value = mock_res
        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='mbs')
        result = resolver.resolve_profiles(formatted_testmodule_mmd,
                                           ('buildroot', 'srpm-buildroot'))
        expected = {
            'buildroot':
                set(['unzip', 'tar', 'cpio', 'gawk', 'gcc', 'xz', 'sed',
                     'findutils', 'util-linux', 'bash', 'info', 'bzip2',
                     'grep', 'redhat-rpm-config', 'fedora-release',
                     'diffutils', 'make', 'patch', 'shadow-utils', 'coreutils',
                     'which', 'rpm-build', 'gzip', 'gcc-c++']),
            'srpm-buildroot':
                set(['shadow-utils', 'redhat-rpm-config', 'rpm-build',
                     'fedora-release', 'fedpkg-minimal', 'gnupg2',
                     'bash'])
        }

        mbs_url = tests.conf.mbs_url
        expected_query = {
            "name": "platform",
            "stream": "f28",
            "version": "3",
            "context": "00000000",
            "verbose": True,
            "order_desc_by": "version",
            "page": 1,
            "per_page": 10,
            "state": "ready"
        }

        mock_session().get.assert_called_once_with(mbs_url, params=expected_query)
        assert result == expected

    @patch("module_build_service.config.Config.system",
           new_callable=PropertyMock, return_value="test")
    @patch("module_build_service.config.Config.mock_resultsdir",
           new_callable=PropertyMock,
           return_value=os.path.join(base_dir, 'staged_data', "local_builds"))
    def test_resolve_profiles_local_module(self, local_builds, conf_system,
                                           formatted_testmodule_mmd):
        tests.clean_database()
        with app.app_context():
            module_build_service.utils.load_local_builds(['platform'])

            resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='mbs')
            result = resolver.resolve_profiles(formatted_testmodule_mmd,
                                               ('buildroot', 'srpm-buildroot'))
            expected = {
                'buildroot':
                    set(['foo']),
                'srpm-buildroot':
                    set(['bar'])
            }
            assert result == expected



import pkg_resources

from module_build_service.builder.base import GenericBuilder

__all__ = [
    GenericBuilder
]

for entrypoint in pkg_resources.iter_entry_points('mbs.builder_backends'):
    GenericBuilder.register_backend_class(entrypoint.load())



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

import mock

import module_build_service.messaging
import module_build_service.scheduler.handlers.repos
import module_build_service.models
from tests import conf, db, app, scheduler_init_data


class TestRepoDone:

    @mock.patch('module_build_service.models.ModuleBuild.from_repo_done_event')
    def test_no_match(self, from_repo_done_event):
        """ Test that when a repo msg hits us and we have no match,
        that we do nothing gracefully.
        """
        scheduler_init_data()
        from_repo_done_event.return_value = None
        msg = module_build_service.messaging.KojiRepoChange(
            'no matches for this...', '2016-some-nonexistent-build')
        module_build_service.scheduler.handlers.repos.done(
            config=conf, session=db.session, msg=msg)

    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.recover_orphaned_artifact', return_value=[])
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.get_average_build_time',
                return_value=0.0)
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.list_tasks_for_components',
                return_value=[])
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.buildroot_ready', return_value=True)
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.get_session')
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.build')
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.buildroot_connect')
    def test_a_single_match(self, connect, build_fn, get_session, ready, list_tasks_fn, mock_gabt,
                            mock_uea):
        """ Test that when a repo msg hits us and we have a single match.
        """
        scheduler_init_data()
        get_session.return_value = mock.Mock(), 'development'
        build_fn.return_value = 1234, 1, '', None

        msg = module_build_service.messaging.KojiRepoChange(
            'some_msg_id', 'module-testmodule-master-20170109091357-7c29193d-build')
        module_build_service.scheduler.handlers.repos.done(
            config=conf, session=db.session, msg=msg)
        build_fn.assert_called_once_with(
            artifact_name='tangerine',
            source=('git://pkgs.fedoraproject.org/rpms/tangerine?'
                    '#fbed359411a1baa08d4a88e0d12d426fbf8f602c'))

    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.recover_orphaned_artifact', return_value=[])
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.get_average_build_time',
                return_value=0.0)
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.list_tasks_for_components',
                return_value=[])
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.buildroot_ready', return_value=True)
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.get_session')
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.build')
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.buildroot_connect')
    def test_a_single_match_build_fail(self, connect, build_fn, config, ready, list_tasks_fn,
                                       mock_gabt, mock_uea):
        """ Test that when a KojiModuleBuilder.build fails, the build is
        marked as failed with proper state_reason.
        """
        scheduler_init_data()
        config.return_value = mock.Mock(), 'development'
        build_fn.return_value = None, 4, 'Failed to submit artifact tangerine to Koji', None

        msg = module_build_service.messaging.KojiRepoChange(
            'some_msg_id', 'module-testmodule-master-20170109091357-7c29193d-build')
        module_build_service.scheduler.handlers.repos.done(
            config=conf, session=db.session, msg=msg)
        build_fn.assert_called_once_with(
            artifact_name='tangerine',
            source=('git://pkgs.fedoraproject.org/rpms/tangerine?'
                    '#fbed359411a1baa08d4a88e0d12d426fbf8f602c'))
        component_build = module_build_service.models.ComponentBuild.query\
            .filter_by(package='tangerine').one()
        assert component_build.state_reason == 'Failed to submit artifact tangerine to Koji'

    @mock.patch('module_build_service.scheduler.handlers.repos.log.info')
    def test_erroneous_regen_repo_received(self, mock_log_info):
        """ Test that when an unexpected KojiRepoRegen message is received, the module doesn't
        complete or go to the next build batch.
        """
        scheduler_init_data(1)
        msg = module_build_service.messaging.KojiRepoChange(
            'some_msg_id', 'module-testmodule-master-20170109091357-7c29193d-build')
        component_build = module_build_service.models.ComponentBuild.query\
            .filter_by(package='tangerine').one()
        component_build.tagged = False
        db.session.add(component_build)
        db.session.commit()
        module_build_service.scheduler.handlers.repos.done(
            config=conf, session=db.session, msg=msg)
        mock_log_info.assert_called_with(
            'Ignoring repo regen, because not all components are tagged.')
        module_build = module_build_service.models.ModuleBuild.query.get(2)
        # Make sure the module build didn't transition since all the components weren't tagged
        assert module_build.state == module_build_service.models.BUILD_STATES['build']

    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.list_tasks_for_components',
                return_value=[])
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.buildroot_ready', return_value=True)
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.get_session')
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.build')
    @mock.patch('module_build_service.builder.KojiModuleBuilder.'
                'KojiModuleBuilder.buildroot_connect')
    @mock.patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
                return_value={'build': [], 'srpm-build': []})
    def test_failed_component_build(self, dbg, connect, build_fn, config, ready, list_tasks_fn):
        """ Test that when a KojiModuleBuilder.build fails, the build is
        marked as failed with proper state_reason.
        """
        with app.app_context():
            scheduler_init_data(3)
            config.return_value = mock.Mock(), 'development'
            build_fn.return_value = None, 4, 'Failed to submit artifact x to Koji', None

            msg = module_build_service.messaging.KojiRepoChange(
                'some_msg_id', 'module-testmodule-master-20170109091357-7c29193d-build')
            module_build_service.scheduler.handlers.repos.done(
                config=conf, session=db.session, msg=msg)
            module_build = module_build_service.models.ModuleBuild.query.get(2)

            assert module_build.state == module_build_service.models.BUILD_STATES["failed"]



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Petr Šabata <contyk@redhat.com>
#            Luboš Kocman <lkocman@redhat.com>


import logging
import os
import koji
import tempfile
import glob
import datetime
import time
import dogpile.cache
import random
import string
import kobo.rpmlib
import threading
try:
    import xmlrpclib
except ImportError:
    import xmlrpc.client as xmlrpclib

import munch
from itertools import chain
from OpenSSL.SSL import SysCallError


from module_build_service import log, conf, models
import module_build_service.scm
import module_build_service.utils
from module_build_service.builder.utils import execute_cmd
from module_build_service.builder.koji_backports import ClientSession as KojiClientSession
from module_build_service.errors import ProgrammingError

from module_build_service.builder.base import GenericBuilder
from module_build_service.builder.KojiContentGenerator import KojiContentGenerator
from module_build_service.utils import get_reusable_components, get_reusable_module
from module_build_service.utils import get_build_arches

logging.basicConfig(level=logging.DEBUG)


def koji_multicall_map(koji_session, koji_session_fnc, list_of_args=None, list_of_kwargs=None):
    """
    Calls the `koji_session_fnc` using Koji multicall feature N times based on the list of
    arguments passed in `list_of_args` and `list_of_kwargs`.
    Returns list of responses sorted the same way as input args/kwargs. In case of error,
    the error message is logged and None is returned.

    For example to get the package ids of "httpd" and "apr" packages:
        ids = koji_multicall_map(session, session.getPackageID, ["httpd", "apr"])
        # ids is now [280, 632]

    :param KojiSessions koji_session: KojiSession to use for multicall.
    :param object koji_session_fnc: Python object representing the KojiSession method to call.
    :param list list_of_args: List of args which are passed to each call of koji_session_fnc.
    :param list list_of_kwargs: List of kwargs which are passed to each call of koji_session_fnc.
    """
    if list_of_args is None and list_of_kwargs is None:
        raise ProgrammingError("One of list_of_args or list_of_kwargs must be set.")

    if (type(list_of_args) not in [type(None), list] or
            type(list_of_kwargs) not in [type(None), list]):
        raise ProgrammingError("list_of_args and list_of_kwargs must be list or None.")

    if list_of_kwargs is None:
        list_of_kwargs = [{}] * len(list_of_args)
    if list_of_args is None:
        list_of_args = [[]] * len(list_of_kwargs)

    if len(list_of_args) != len(list_of_kwargs):
        raise ProgrammingError("Length of list_of_args and list_of_kwargs must be the same.")

    koji_session.multicall = True
    for args, kwargs in zip(list_of_args, list_of_kwargs):
        if type(args) != list:
            args = [args]
        if type(kwargs) != dict:
            raise ProgrammingError("Every item in list_of_kwargs must be a dict")
        koji_session_fnc(*args, **kwargs)

    try:
        responses = koji_session.multiCall(strict=True)
    except Exception:
        log.exception("Exception raised for multicall of method %r with args %r, %r:",
                      koji_session_fnc, args, kwargs)
        return None

    if not responses:
        log.error("Koji did not return response for multicall of %r", koji_session_fnc)
        return None
    if type(responses) != list:
        log.error("Fault element was returned for multicall of method %r: %r",
                  koji_session_fnc, responses)
        return None

    results = []

    # For the response specification, see
    # https://web.archive.org/web/20060624230303/http://www.xmlrpc.com/discuss/msgReader$1208?mode=topic
    # Relevant part of this:
    # Multicall returns an array of responses. There will be one response for each call in
    # the original array. The result will either be a one-item array containing the result value,
    # or a struct of the form found inside the standard <fault> element.
    for response, args, kwargs in zip(responses, list_of_args, list_of_kwargs):
        if type(response) == list:
            if not response:
                log.error("Empty list returned for multicall of method %r with args %r, %r",
                          koji_session_fnc, args, kwargs)
                return None
            results.append(response[0])
        else:
            log.error("Unexpected data returned for multicall of method %r with args %r, %r: %r",
                      koji_session_fnc, args, kwargs, response)
            return None

    return results


@module_build_service.utils.retry(wait_on=(xmlrpclib.ProtocolError, koji.GenericError))
def koji_retrying_multicall_map(*args, **kwargs):
    """
    Retrying version of koji_multicall_map. This tries to retry the Koji call
    in case of koji.GenericError or xmlrpclib.ProtocolError.

    Please refer to koji_multicall_map for further specification of arguments.
    """
    return koji_multicall_map(*args, **kwargs)


class KojiModuleBuilder(GenericBuilder):
    """ Koji specific builder class """

    backend = "koji"
    _build_lock = threading.Lock()
    region = dogpile.cache.make_region().configure('dogpile.cache.memory')

    @module_build_service.utils.validate_koji_tag('tag_name')
    def __init__(self, owner, module, config, tag_name, components):
        """
        :param owner: a string representing who kicked off the builds
        :param module: module_build_service.models.ModuleBuild instance.
        :param config: module_build_service.config.Config instance
        :param tag_name: name of tag for given module
        """
        self.owner = owner
        self.module_str = module.name
        self.module = module
        self.mmd = module.mmd()
        self.config = config
        self.tag_name = tag_name
        self.__prep = False
        log.debug("Using koji profile %r" % config.koji_profile)
        log.debug("Using koji_config: %s" % config.koji_config)

        self.koji_session = self.get_session(config)
        self.arches = get_build_arches(self.mmd, self.config)

        if not self.arches:
            raise ValueError("No arches specified in the config.")

        # These eventually get populated by calling _connect and __prep is set to True
        self.module_tag = None  # string
        # A dict containing tag info returned from Koji API getTag.
        self.module_build_tag = None
        self.module_target = None  # A koji target dict

        self.build_priority = config.koji_build_priority
        self.components = components

    def __repr__(self):
        return "<KojiModuleBuilder module: %s, tag: %s>" % (
            self.module_str, self.tag_name)

    @region.cache_on_arguments()
    def getPerms(self):
        return dict([(p['name'], p['id']) for p in self.koji_session.getAllPerms()])

    @module_build_service.utils.retry(wait_on=(IOError, koji.GenericError))
    def buildroot_ready(self, artifacts=None):
        """
        :param artifacts=None - list of nvrs
        Returns True or False if the given artifacts are in the build root.
        """
        assert self.module_target, "Invalid build target"

        tag_id = self.module_target['build_tag']
        repo = self.koji_session.getRepo(tag_id)
        builds = [self.koji_session.getBuild(a, strict=True) for a in artifacts or []]
        log.info("%r checking buildroot readiness for "
                 "repo: %r, tag_id: %r, artifacts: %r, builds: %r" % (
                     self, repo, tag_id, artifacts, builds))

        if not repo:
            log.info("Repo is not generated yet, buildroot is not ready yet.")
            return False

        ready = bool(koji.util.checkForBuilds(
            self.koji_session,
            tag_id,
            builds,
            repo['create_event'],
            latest=True,
        ))
        if ready:
            log.info("%r buildroot is ready" % self)
        else:
            log.info("%r buildroot is not yet ready.. wait." % self)
        return ready

    @staticmethod
    def _get_filtered_rpms_on_self_dep(module_build, filtered_rpms_of_dep):
        # filtered_rpms will contain the NVRs of non-reusable component's RPMs
        filtered_rpms = list(set(filtered_rpms_of_dep))
        with models.make_session(conf) as db_session:
            # Get a module build that can be reused, which will likely be the
            # build dep that is used since it relies on itself
            reusable_module = get_reusable_module(db_session, module_build)
            if not reusable_module:
                return filtered_rpms
            koji_session = KojiModuleBuilder.get_session(conf, login=False)
            # Get all the RPMs and builds of the reusable module in Koji
            rpms, builds = koji_session.listTaggedRPMS(reusable_module.koji_tag, latest=True)
            # Convert the list to a dict where each key is the build_id
            builds = {build['build_id']: build for build in builds}
            # Create a mapping of package (SRPM) to the RPMs in NVR format
            package_to_rpms = {}
            for rpm in rpms:
                package = builds[rpm['build_id']]['name']
                if package not in package_to_rpms:
                    package_to_rpms[package] = []
                package_to_rpms[package].append(kobo.rpmlib.make_nvr(rpm))

            components_in_module = [c.package for c in module_build.component_builds]
            reusable_components = get_reusable_components(
                db_session, module_build, components_in_module,
                previous_module_build=reusable_module)
            # Loop through all the reusable components to find if any of their RPMs are
            # being filtered
            for reusable_component in reusable_components:
                # reusable_component will be None if the component can't be reused
                if not reusable_component:
                    continue
                # We must get the component name from the NVR and not from
                # reusable_component.package because macros such as those used
                # by SCLs can change the name of the underlying build
                component_name = kobo.rpmlib.parse_nvr(reusable_component.nvr)['name']

                if component_name not in package_to_rpms:
                    continue

                # Loop through the RPMs associated with the reusable component
                for nvr in package_to_rpms[component_name]:
                    parsed_nvr = kobo.rpmlib.parse_nvr(nvr)
                    # Don't compare with the epoch
                    parsed_nvr['epoch'] = None
                    # Loop through all the filtered RPMs to find a match with the reusable
                    # component's RPMs.
                    for nvr2 in list(filtered_rpms):
                        parsed_nvr2 = kobo.rpmlib.parse_nvr(nvr2)
                        # Don't compare with the epoch
                        parsed_nvr2['epoch'] = None
                        # Only remove the filter if we are going to reuse a component with
                        # the same exact NVR
                        if parsed_nvr == parsed_nvr2:
                            filtered_rpms.remove(nvr2)
                            # Since filtered_rpms was cast to a set and then back
                            # to a list above, we know there won't be duplicate RPMS,
                            # so we can just break here.
                            break
        return filtered_rpms

    @staticmethod
    def format_conflicts_line(nevr):
        """Helper method to format a Conflicts line with a RPM N-E:V-R"""
        parsed_nvr = kobo.rpmlib.parse_nvr(nevr)
        return "Conflicts: {name} = {epoch}:{version}-{release}".format(**parsed_nvr)

    @staticmethod
    def get_disttag_srpm(disttag, module_build):

        # Taken from Karsten's create-distmacro-pkg.sh
        # - however removed any provides to system-release/redhat-release

        name = 'module-build-macros'
        version = "0.1"
        release = "1"
        today = datetime.date.today().strftime('%a %b %d %Y')
        mmd = module_build.mmd()

        # Generate "Conflicts: name = version-release". This is workaround for
        # Koji build system, because it does not filter out RPMs from the
        # build-requires based on their "mmd.filter.rpms". So we set the
        # module-build-macros to conflict with these filtered RPMs to ensure
        # they won't be installed to buildroot.
        filter_conflicts = []
        for req_name, req_data in mmd.get_xmd()["mbs"]["buildrequires"].items():
            if req_data["filtered_rpms"]:
                filter_conflicts.append("# Filtered rpms from %s module:" % req_name)
            # Check if the module depends on itself
            if req_name == module_build.name:
                filtered_rpms = KojiModuleBuilder._get_filtered_rpms_on_self_dep(
                    module_build, req_data["filtered_rpms"])
            else:
                filtered_rpms = req_data["filtered_rpms"]
            filter_conflicts.extend(map(
                KojiModuleBuilder.format_conflicts_line, filtered_rpms))

            if req_name in conf.base_module_names and 'ursine_rpms' in req_data:
                comments = (
                    '# Filter out RPMs from stream collision modules found from ursine content'
                    ' for base module {}:'.format(req_name),
                    '# ' + ', '.join(req_data['stream_collision_modules']),
                )
                filter_conflicts.extend(chain(
                    comments,
                    map(KojiModuleBuilder.format_conflicts_line, req_data['ursine_rpms'])
                ))

        spec_content = """
%global dist {disttag}
%global disttag module({module_name}:{module_stream}:{module_version}:{module_context})
%global modularitylabel {module_name}:{module_stream}:{module_version}:{module_context}
%global _module_name {module_name}
%global _module_stream {module_stream}
%global _module_version {module_version}
%global _module_context {module_context}

Name:       {name}
Version:    {version}
Release:    {release}%dist
Summary:    Package containing macros required to build generic module
BuildArch:  noarch

Group:      System Environment/Base
License:    MIT
URL:        http://fedoraproject.org

Source1:    macros.modules

{filter_conflicts}

%description
This package is used for building modules with a different dist tag.
It provides a file /usr/lib/rpm/macros.d/macro.modules and gets read
after macro.dist, thus overwriting macros of macro.dist like %%dist
It should NEVER be installed on any system as it will really mess up
 updates, builds, ....


%build

%install
mkdir -p %buildroot/etc/rpm 2>/dev/null |:
cp %SOURCE1 %buildroot/etc/rpm/macros.zz-modules
chmod 644 %buildroot/etc/rpm/macros.zz-modules


%files
/etc/rpm/macros.zz-modules



%changelog
* {today} Fedora-Modularity - {version}-{release}{disttag}
- autogenerated macro by Module Build Service (MBS)
""".format(disttag=disttag, today=today, name=name, version=version,
           release=release,
           module_name=module_build.name,
           module_stream=module_build.stream,
           module_version=module_build.version,
           module_context=module_build.context,
           filter_conflicts='\n'.join(filter_conflicts))

        modulemd_macros = ""
        rpm_buildopts = mmd.get_rpm_buildopts()
        if rpm_buildopts:
            modulemd_macros = rpm_buildopts.get('macros')

        macros_content = """

# General macros set by MBS

%dist {disttag}
%disttag module({module_name}:{module_stream}:{module_version}:{module_context})
%modularitylabel {module_name}:{module_stream}:{module_version}:{module_context}
%_module_build 1
%_module_name {module_name}
%_module_stream {module_stream}
%_module_version {module_version}
%_module_context {module_context}

# Macros set by module author:

{modulemd_macros}
""".format(disttag=disttag, module_name=module_build.name,
           module_stream=module_build.stream,
           module_version=module_build.version,
           module_context=module_build.context,
           modulemd_macros=modulemd_macros)

        td = tempfile.mkdtemp(prefix="module_build_service-build-macros")
        fd = open(os.path.join(td, "%s.spec" % name), "w")
        fd.write(spec_content)
        fd.close()
        sources_dir = os.path.join(td, "SOURCES")
        os.mkdir(sources_dir)
        fd = open(os.path.join(sources_dir, "macros.modules"), "w")
        fd.write(macros_content)
        fd.close()
        log.debug("Building %s.spec" % name)

        # We are not interested in the rpmbuild stdout...
        null_fd = open(os.devnull, 'w')
        execute_cmd(['rpmbuild', '-bs', '%s.spec' % name,
                     '--define', '_topdir %s' % td,
                     '--define', '_sourcedir %s' % sources_dir],
                    cwd=td, stdout=null_fd)
        null_fd.close()
        sdir = os.path.join(td, "SRPMS")
        srpm_paths = glob.glob("%s/*.src.rpm" % sdir)
        assert len(srpm_paths) == 1, "Expected exactly 1 srpm in %s. Got %s" % (sdir, srpm_paths)

        log.debug("Wrote srpm into %s" % srpm_paths[0])
        return srpm_paths[0]

    @staticmethod
    @module_build_service.utils.retry(wait_on=(xmlrpclib.ProtocolError, koji.GenericError))
    def get_session(config, login=True):
        koji_config = munch.Munch(koji.read_config(
            profile_name=config.koji_profile,
            user_config=config.koji_config,
        ))
        # Timeout after 10 minutes.  The default is 12 hours.
        koji_config["timeout"] = 60 * 10

        address = koji_config.server
        log.info("Connecting to koji %r.", address)
        koji_session = KojiClientSession(address, opts=koji_config)

        if not login:
            return koji_session

        authtype = koji_config.authtype
        log.info("Authenticate session with %r.", authtype)
        if authtype == "kerberos":
            try:
                import krbV
            except ImportError:
                raise RuntimeError(
                    "python-krbV must be installed to authenticate with Koji using Kerberos")
            keytab = getattr(config, "krb_keytab", None)
            principal = getattr(config, "krb_principal", None)
            if not keytab and principal:
                raise ValueError(
                    "The Kerberos keytab and principal aren't set for Koji authentication")
            log.debug("  keytab: %r, principal: %r" % (keytab, principal))
            # We want to create a context per thread to avoid Kerberos cache corruption
            ctx = krbV.Context()
            # We want to use the thread keyring for the ccache to ensure we have one cache per
            # thread to avoid Kerberos cache corruption
            ccache = "KEYRING:thread:mbs"
            koji_session.krb_login(principal=principal, keytab=keytab, ctx=ctx, ccache=ccache)
        elif authtype == "ssl":
            koji_session.ssl_login(
                os.path.expanduser(koji_config.cert),
                None,
                os.path.expanduser(koji_config.serverca)
            )
        else:
            raise ValueError("Unrecognized koji authtype %r" % authtype)

        return koji_session

    def buildroot_connect(self, groups):
        log.info("%r connecting buildroot." % self)

        # Check if the build_tag exists, because there are Koji calls later which must be called
        # only if we are creating the build_tag for first time.
        build_tag_exists = self.koji_session.getTag(self.tag_name + "-build")

        # Create or update individual tags
        # the main tag needs arches so pungi can dump it
        self.module_tag = self._koji_create_tag(
            self.tag_name, self.arches, perm="admin")
        self.module_build_tag = self._koji_create_tag(
            self.tag_name + "-build", self.arches, perm="admin")

        self._koji_whitelist_packages(
            self.mmd.props.buildopts.props.rpm_whitelist or self.components)

        # If we have just created the build tag in this buildroot_connect call, block all
        # the components in `blocked_packages` list. We want to do that just once, because
        # there might be some unblocked packages later and we would block them again...
        if not build_tag_exists:
            xmd = self.mmd.get_xmd()
            if "mbs_options" in xmd.keys() and "blocked_packages" in xmd["mbs_options"].keys():
                self._koji_block_packages(xmd["mbs_options"]["blocked_packages"])

        @module_build_service.utils.retry(wait_on=SysCallError, interval=5)
        def add_groups():
            return self._koji_add_groups_to_tag(
                dest_tag=self.module_build_tag,
                groups=groups,
            )
        add_groups()

        # Koji targets can only be 50 characters long, but the generate_koji_tag function
        # checks the length with '-build' at the end, but we know we will never append '-build',
        # so we can safely have the name check be more characters
        target_length = 50 + len('-build')
        target = module_build_service.utils.generate_koji_tag(
            self.module.name, self.module.stream, self.module.version, self.module.context,
            target_length)
        # Add main build target.
        self.module_target = self._koji_add_target(target, self.module_build_tag, self.module_tag)

        self.__prep = True
        log.info("%r buildroot sucessfully connected." % self)

    def buildroot_add_repos(self, dependencies):
        log.info("%r adding deps on %r" % (self, dependencies))
        self._koji_add_many_tag_inheritance(self.module_build_tag, dependencies)

    def _get_tagged_nvrs(self, tag):
        """
        Returns set of NVR strings tagged in tag `tag`.
        """
        tagged = self.koji_session.listTagged(tag)
        tagged_nvrs = set(build["nvr"] for build in tagged)
        return tagged_nvrs

    def buildroot_add_artifacts(self, artifacts, install=False):
        """
        :param artifacts - list of artifacts to add to buildroot
        :param install=False - force install artifact (if it's not dragged in as dependency)

        This method is safe to call multiple times.
        """
        log.info("%r adding artifacts %r" % (self, artifacts))
        build_tag = self._get_tag(self.module_build_tag)['id']

        xmd = self.mmd.get_xmd()
        if "mbs_options" in xmd.keys() and "blocked_packages" in xmd["mbs_options"].keys():
            packages = [kobo.rpmlib.parse_nvr(nvr)["name"] for nvr in artifacts]
            packages = [package for package in packages
                        if package in xmd["mbs_options"]["blocked_packages"]]
            if packages:
                self._koji_unblock_packages(packages)

        tagged_nvrs = self._get_tagged_nvrs(self.module_build_tag['name'])

        self.koji_session.multicall = True
        for nvr in artifacts:
            if nvr in tagged_nvrs:
                continue

            log.info("%r tagging %r into %r" % (self, nvr, build_tag))
            self.koji_session.tagBuild(build_tag, nvr)

            if not install:
                continue

            for group in ('srpm-build', 'build'):
                name = kobo.rpmlib.parse_nvr(nvr)['name']
                log.info("%r adding %s to group %s" % (self, name, group))
                self.koji_session.groupPackageListAdd(build_tag, group, name)
        self.koji_session.multiCall(strict=True)

    def tag_artifacts(self, artifacts, dest_tag=True):
        """ Tag the provided artifacts to the module tag
        :param artifacts: a list of NVRs to tag
        :kwarg dest_tag: a boolean determining if the destination or build tag should be used
        :return: None
        """
        if dest_tag:
            tag = self._get_tag(self.module_tag)['id']
            tagged_nvrs = self._get_tagged_nvrs(self.module_tag['name'])
        else:
            tag = self._get_tag(self.module_build_tag)['id']
            tagged_nvrs = self._get_tagged_nvrs(self.module_build_tag['name'])

        self.koji_session.multicall = True
        for nvr in artifacts:
            if nvr in tagged_nvrs:
                continue

            log.info("%r tagging %r into %r" % (self, nvr, tag))
            self.koji_session.tagBuild(tag, nvr)
        self.koji_session.multiCall(strict=True)

    def untag_artifacts(self, artifacts):
        """ Untag the provided artifacts from the module destination and build tag
        :param artifacts: a list of NVRs to untag
        :return: None
        """
        build_tag_name = self.tag_name + '-build'
        dest_tag = self._get_tag(self.tag_name, strict=False)
        build_tag = self._get_tag(build_tag_name, strict=False)
        # Get the NVRs in the tags to make sure the builds exist and they're tagged before
        # untagging them
        if dest_tag:
            dest_tagged_nvrs = self._get_tagged_nvrs(dest_tag['name'])
        else:
            log.info('The tag "{0}" doesn\'t exist'.format(self.tag_name))
            dest_tagged_nvrs = []
        if build_tag:
            build_tagged_nvrs = self._get_tagged_nvrs(build_tag['name'])
        else:
            log.info('The tag "{0}" doesn\'t exist'.format(build_tag_name))
            build_tagged_nvrs = []

        # If there is nothing to untag, then just return
        if not dest_tagged_nvrs and not build_tagged_nvrs:
            return

        self.koji_session.multicall = True
        for nvr in artifacts:
            if nvr in dest_tagged_nvrs:
                log.info("%r untagging %r from %r" % (self, nvr, dest_tag['id']))
                self.koji_session.untagBuild(dest_tag['id'], nvr)
            if nvr in build_tagged_nvrs:
                log.info("%r untagging %r from %r" % (self, nvr, build_tag['id']))
                self.koji_session.untagBuild(build_tag['id'], nvr)
        self.koji_session.multiCall(strict=True)

    def wait_task(self, task_id):
        """
        :param task_id
        :return - task result object
        """

        log.info("Waiting for task_id=%s to finish" % task_id)

        timeout = 60 * 60  # 60 minutes

        @module_build_service.utils.retry(timeout=timeout, wait_on=koji.GenericError)
        def get_result():
            log.debug("Waiting for task_id=%s to finish" % task_id)
            task = self.koji_session.getTaskResult(task_id)
            log.info("Done waiting for task_id=%s to finish" % task_id)
            return task

        return get_result()

    def recover_orphaned_artifact(self, component_build):
        """
        Searches for a complete build of an artifact belonging to the module and sets the
        component_build in the MBS database to the found build. This usually returns nothing since
        these builds should *not* exist.
        :param component_build: a ComponentBuild object
        :return: a list of msgs that MBS needs to process
        """
        opts = {'latest': True, 'package': component_build.package, 'inherit': False}
        build_tagged = self.koji_session.listTagged(self.module_build_tag['name'], **opts)
        dest_tagged = None
        # Only check the destination tag if the component is not a build_time_only component
        if not component_build.build_time_only:
            dest_tagged = self.koji_session.listTagged(self.module_tag['name'], **opts)
        for rv in [build_tagged, dest_tagged]:
            if rv and len(rv) != 1:
                raise ValueError("Expected exactly one item in list. Got %s" % rv)

        build = None
        if build_tagged:
            build = build_tagged[0]
        elif dest_tagged:
            build = dest_tagged[0]

        if not build:
            # If the build cannot be found in the tags, it may be untagged as a result
            # of some earlier inconsistent situation. Let's find the task_info
            # based on the list of untagged builds
            release = module_build_service.utils.get_rpm_release(self.module)
            untagged = self.koji_session.untaggedBuilds(name=component_build.package)
            for untagged_build in untagged:
                if untagged_build["release"].endswith(release):
                    nvr = "{name}-{version}-{release}".format(**untagged_build)
                    build = self.koji_session.getBuild(nvr)
                    break
        further_work = []
        # If the build doesn't exist, then return
        if not build:
            return further_work

        # Start setting up MBS' database to use the existing build
        log.info('Skipping build of "{0}" since it already exists.'.format(build['nvr']))
        # Set it to COMPLETE so it doesn't count towards the concurrent component threshold
        component_build.state = koji.BUILD_STATES['COMPLETE']
        component_build.nvr = build['nvr']
        component_build.task_id = build['task_id']
        component_build.state_reason = 'Found existing build'
        nvr_dict = kobo.rpmlib.parse_nvr(component_build.nvr)
        # Trigger a completed build message
        further_work.append(module_build_service.messaging.KojiBuildChange(
            'recover_orphaned_artifact: fake message', build['build_id'],
            build['task_id'], koji.BUILD_STATES['COMPLETE'], component_build.package,
            nvr_dict['version'], nvr_dict['release'], component_build.module_build.id))

        component_tagged_in = []
        if build_tagged:
            component_tagged_in.append(self.module_build_tag['name'])
        else:
            # Tag it in the build tag if it's not there
            self.tag_artifacts([component_build.nvr], dest_tag=False)
        if dest_tagged:
            component_tagged_in.append(self.module_tag['name'])
        for tag in component_tagged_in:
            log.info('The build being skipped isn\'t tagged in the "{0}" tag. Will send a '
                     'message to the tag handler'.format(tag))
            further_work.append(module_build_service.messaging.KojiTagChange(
                'recover_orphaned_artifact: fake message', tag, component_build.package,
                component_build.nvr))
        return further_work

    def build(self, artifact_name, source):
        """
        :param artifact_name: a string of the name of the artifact
        :param source: a string of the scmurl to the spec repository
        :return: 4-tuple of the form (koji build task id, state, reason, nvr)
        """

        # TODO: If we are sure that this method is thread-safe, we can just
        # remove _build_lock locking.
        with KojiModuleBuilder._build_lock:
            # This code supposes that artifact_name can be built within the component
            # Taken from /usr/bin/koji
            def _unique_path(prefix):
                """
                Create a unique path fragment by appending a path component
                to prefix.  The path component will consist of a string of letter and numbers
                that is unlikely to be a duplicate, but is not guaranteed to be unique.
                """
                # Use time() in the dirname to provide a little more information when
                # browsing the filesystem.
                # For some reason repr(time.time()) includes 4 or 5
                # more digits of precision than str(time.time())
                # Unnamed Engineer: Guido v. R., I am disappoint
                return '%s/%r.%s' % (prefix, time.time(),
                                     ''.join([random.choice(string.ascii_letters)
                                              for i in range(8)]))

            if not self.__prep:
                raise RuntimeError("Buildroot is not prep-ed")

            self._koji_whitelist_packages([artifact_name])
            if '://' not in source:
                # treat source as an srpm and upload it
                serverdir = _unique_path('cli-build')
                callback = None
                self.koji_session.uploadWrapper(source, serverdir, callback=callback)
                source = "%s/%s" % (serverdir, os.path.basename(source))

            # When "koji_build_macros_target" is set, we build the
            # module-build-macros in this target instead of the self.module_target.
            # The reason is that it is faster to build this RPM in
            # already existing shared target, because Koji does not need to do
            # repo-regen.
            if (artifact_name == "module-build-macros" and
               self.config.koji_build_macros_target):
                module_target = self.config.koji_build_macros_target
            else:
                module_target = self.module_target['name']

            build_opts = {
                "skip_tag": True,
                "mbs_artifact_name": artifact_name,
                "mbs_module_target": module_target
            }

            # disabled by default, wouldn't work until Koji issue #1158 is done
            if conf.allow_arch_override:
                build_opts['arch_override'] = \
                    self.mmd.get_rpm_components()[artifact_name].get_arches().get()

            task_id = self.koji_session.build(source, module_target, build_opts,
                                              priority=self.build_priority)
            log.info("submitted build of %s (task_id=%s), via %s" % (
                source, task_id, self))
            if task_id:
                state = koji.BUILD_STATES['BUILDING']
                reason = "Submitted %s to Koji" % (artifact_name)
            else:
                state = koji.BUILD_STATES['FAILED']
                reason = "Failed to submit artifact %s to Koji" % (artifact_name)
            return task_id, state, reason, None

    def cancel_build(self, task_id):
        try:
            self.koji_session.cancelTask(task_id)
        except Exception as error:
            log.error('Failed to cancel task ID {0} in Koji. The error '
                      'message was: {1}'.format(task_id, str(error)))

    @classmethod
    def repo_from_tag(cls, config, tag_name, arch):
        """
        :param config: instance of module_build_service.config.Config
        :param tag_name: Tag for which the repository is returned
        :param arch: Architecture for which the repository is returned

        Returns URL of repository containing the built artifacts for
        the tag with particular name and architecture.
        """
        return "%s/%s/latest/%s" % (config.koji_repository_url, tag_name, arch)

    @module_build_service.utils.validate_koji_tag('tag', post='')
    def _get_tag(self, tag, strict=True):
        if isinstance(tag, dict):
            tag = tag['name']
        taginfo = self.koji_session.getTag(tag)
        if not taginfo:
            if strict:
                raise SystemError("Unknown tag: %s" % tag)
        return taginfo

    @module_build_service.utils.validate_koji_tag(['tag_name'], post='')
    def _koji_add_many_tag_inheritance(self, tag_name, parent_tags):
        tag = self._get_tag(tag_name)
        # highest priority num is at the end
        inheritance_data = sorted(self.koji_session.getInheritanceData(tag['name']) or
                                  [], key=lambda k: k['priority'])
        # Set initial priority to last record in inheritance data or 0
        priority = 0
        if inheritance_data:
            priority = inheritance_data[-1]['priority'] + 10

        def record_exists(parent_id, data):
            for item in data:
                if parent_id == item['parent_id']:
                    return True
            return False

        for parent in parent_tags:  # We expect that they're sorted
            parent = self._get_tag(parent)
            if record_exists(parent['id'], inheritance_data):
                continue

            parent_data = {}
            parent_data['parent_id'] = parent['id']
            parent_data['priority'] = priority
            parent_data['maxdepth'] = None
            parent_data['intransitive'] = False
            parent_data['noconfig'] = False
            parent_data['pkg_filter'] = ''
            inheritance_data.append(parent_data)
            priority += 10

        if inheritance_data:
            self.koji_session.setInheritanceData(tag['id'], inheritance_data)

    @module_build_service.utils.validate_koji_tag('dest_tag')
    def _koji_add_groups_to_tag(self, dest_tag, groups):
        """Add groups to a tag as well as packages listed by group

        :param dest_tag: the tag groups will be added to. This argument could
            be a string representing tag name, or a mapping containing tag info
            which must have name at least.
        :type dest_tag: str or dict
        :param dict groups: A dict ``{'group' : [package, ...]}``. If one of
            the groups has been added to the tag, the group is skipped.
        """
        log.debug("Adding groups=%s to tag=%s" % (list(groups), dest_tag))
        if groups and not isinstance(groups, dict):
            raise ValueError("Expected dict {'group' : [str(package1), ...]")
        dest_tag = self._get_tag(dest_tag)['name']
        existing_groups = dict([(p['name'], p['group_id'])
                                for p
                                in self.koji_session.getTagGroups(dest_tag, inherit=False)
                                ])

        for group, packages in groups.items():
            group_id = existing_groups.get(group, None)
            if group_id is not None:
                log.debug("Group %s already exists for tag %s. Skipping creation."
                          % (group, dest_tag))
                continue

            self.koji_session.groupListAdd(dest_tag, group)
            log.debug("Adding %d packages into group=%s tag=%s" % (len(packages), group, dest_tag))

            # This doesn't fail in case that it's already present in the group. This should be safe
            for pkg in packages:
                self.koji_session.groupPackageListAdd(dest_tag, group, pkg)

    @module_build_service.utils.validate_koji_tag('tag_name')
    def _koji_create_tag(self, tag_name, arches=None, perm=None):
        """Create a tag in Koji

        This call is safe to call multiple times.

        :param str tag_name: name of koji tag
        :kwarg list arches: list of architectures for the tag
        :kwarg str perm: permissions for the tag (used in lock-tag)
        :return: a mapping containing raw tag info returned from Koji.
        :rtype: dict
        """

        log.debug("Ensuring existence of tag='%s'." % tag_name)
        taginfo = self.koji_session.getTag(tag_name)

        if not taginfo:
            self.koji_session.createTag(tag_name)
            taginfo = self._get_tag(tag_name)

        opts = {}
        if arches:
            if not isinstance(arches, list):
                raise ValueError("Expected list or None on input got %s" % type(arches))

            current_arches = []
            if taginfo['arches']:  # None if none
                current_arches = taginfo['arches'].split()  # string separated by empty spaces

            if set(arches) != set(current_arches):
                opts['arches'] = " ".join(arches)

        if perm:
            if taginfo['locked']:
                raise SystemError("Tag %s: master lock already set. Can't edit tag"
                                  % taginfo['name'])

            perm_ids = self.getPerms()

            if perm not in perm_ids:
                raise ValueError("Unknown permissions %s" % perm)

            perm_id = perm_ids[perm]
            if taginfo['perm'] not in (perm_id, perm):  # check either id or the string
                opts['perm'] = perm_id

        opts['extra'] = {
            'mock.package_manager': 'dnf',
            # This is needed to include all the Koji builds (and therefore
            # all the packages) from all inherited tags into this tag.
            # See https://pagure.io/koji/issue/588 and
            # https://pagure.io/fm-orchestrator/issue/660 for background.
            'repo_include_all': True,
        }

        xmd = self.mmd.get_xmd()
        if "mbs_options" in xmd.keys() and "repo_include_all" in xmd["mbs_options"].keys():
            opts['extra']['repo_include_all'] = xmd["mbs_options"]["repo_include_all"]

        # edit tag with opts
        self.koji_session.editTag2(tag_name, **opts)
        return self._get_tag(tag_name)  # Return up2date taginfo

    def _koji_whitelist_packages(self, packages, tags=None):
        if not tags:
            tags = [self.module_tag, self.module_build_tag]

        # This will help with potential resubmitting of failed builds
        pkglists = {}
        for tag in tags:
            pkglists[tag['id']] = dict([(p['package_name'], p['package_id'])
                                        for p in self.koji_session.listPackages(tagID=tag['id'])])

        self.koji_session.multicall = True
        for tag in tags:
            pkglist = pkglists[tag['id']]
            for package in packages:
                if pkglist.get(package, None):
                    log.debug("%s Package %s is already whitelisted." % (self, package))
                    continue

                self.koji_session.packageListAdd(tag['name'], package, self.owner)
        self.koji_session.multiCall(strict=True)

    def _koji_block_packages(self, packages):
        """
        Blocks the `packages` for the module_build_tag.
        """
        log.info("Blocking packages in tag %s: %r", self.module_build_tag["name"], packages)
        args = [[self.module_build_tag["name"], package] for package in packages]
        koji_multicall_map(self.koji_session, self.koji_session.packageListBlock, args)

    def _koji_unblock_packages(self, packages):
        """
        Unblocks the `packages` for the module_build_tag.
        """
        log.info("Unblocking packages in tag %s: %r", self.module_build_tag["name"], packages)
        args = [[self.module_build_tag["name"], package] for package in packages]
        koji_multicall_map(self.koji_session, self.koji_session.packageListUnblock, args)

    @module_build_service.utils.validate_koji_tag(['build_tag', 'dest_tag'])
    def _koji_add_target(self, name, build_tag, dest_tag):
        """Add build target if it doesn't exist or validate the existing one

        This call is safe to call multiple times. Raises SystemError() if the existing target
        doesn't match params. The reason not to touch existing target, is that we don't want to
        accidentally alter a target which was already used to build some artifacts.

        :param str name: target name.
        :param dict build_tag: build tag info, which must have name at least.
        :param dict dest_tag: dest tag info, which must have name at least.
        :return: a mapping containing raw build target info returned from Koji.
            If a new build target is created, it is the new one. Otherwise,
            existing build target is returned.
        :rtype: dict
        :raises SystemError: if existing build target does not have build_tag
            name or dest_tag name.
        """
        build_tag = self._get_tag(build_tag)
        dest_tag = self._get_tag(dest_tag)
        target_info = self.koji_session.getBuildTarget(name)

        barches = build_tag.get("arches", None)
        assert barches, "Build tag %s has no arches defined." % build_tag['name']

        if not target_info:
            target_info = self.koji_session.createBuildTarget(name, build_tag['name'],
                                                              dest_tag['name'])

        else:  # verify whether build and destination tag matches
            if build_tag['name'] != target_info['build_tag_name']:
                raise SystemError(("Target references unexpected build_tag_name. "
                                   "Got '%s', expected '%s'. Please contact administrator.")
                                  % (target_info['build_tag_name'], build_tag['name']))
            if dest_tag['name'] != target_info['dest_tag_name']:
                raise SystemError(("Target references unexpected dest_tag_name. "
                                   "Got '%s', expected '%s'. Please contact administrator.")
                                  % (target_info['dest_tag_name'], dest_tag['name']))

        return self.koji_session.getBuildTarget(name)

    def list_tasks_for_components(self, component_builds=None, state='active'):
        """
        :param component_builds: list of component builds which we want to check
        :param state: limit the check only for Koji tasks in the given state
        :return: list of Koji tasks

        List Koji tasks ('active' by default) for component builds.
        """

        component_builds = component_builds or []
        if state == 'active':
            states = [koji.TASK_STATES['FREE'],
                      koji.TASK_STATES['OPEN'],
                      koji.TASK_STATES['ASSIGNED']]
        elif state.upper() in koji.TASK_STATES:
            states = [koji.TASK_STATES[state.upper()]]
        else:
            raise ValueError("State {} is not valid within Koji task states."
                             .format(state))

        tasks = []
        for task in self.koji_session.listTasks(opts={'state': states,
                                                      'decode': True,
                                                      'method': 'build'}):
            task_opts = task['request'][-1]
            assert isinstance(task_opts, dict), "Task options shall be a dict."
            if 'scratch' in task_opts and task_opts['scratch']:
                continue
            if 'mbs_artifact_name' not in task_opts:
                task_opts['mbs_artifact_name'] = None
            if 'mbs_module_target' not in task_opts:
                task_opts['mbs_module_target'] = None
            for c in component_builds:
                # TODO: https://pagure.io/fm-orchestrator/issue/397
                # Subj: Do not mix target/tag when looking for component builds
                if (c.package == task_opts['mbs_artifact_name'] and
                   c.module_build.koji_tag == task_opts['mbs_module_target']):
                    tasks.append(task)

        return tasks

    @classmethod
    def get_average_build_time(cls, component):
        """
        Get the average build time of the component from Koji
        :param component: a ComponentBuild object
        :return: a float of the average build time in seconds
        """
        # If the component has not been built before, then None is returned. Instead, let's
        # return 0.0 so the type is consistent
        koji_session = KojiModuleBuilder.get_session(conf, login=False)
        return koji_session.getAverageBuildDuration(component) or 0.0

    @classmethod
    def get_build_weights(cls, components):
        """
        Returns a dict with component name as a key and float number
        representing the overall Koji weight of a component build.
        The weight is sum of weights of all tasks in a previously done modular
        build of a component.

        :param list components: List of component names.
        :rtype: dict
        :return: {component_name: weight_as_float, ...}
        """

        koji_session = KojiModuleBuilder.get_session(conf)

        # Get our own userID, so we can limit the builds to only modular builds
        user_info = koji_session.getLoggedInUser()
        if not user_info or "id" not in user_info:
            log.warning("Koji.getLoggedInUser() failed while getting build weight.")
            return cls.compute_weights_from_build_time(components)
        mbs_user_id = user_info["id"]

        # Get the Koji PackageID for every component in single Koji call.
        # If some package does not exist in Koji, component_ids will be None.
        component_ids = koji_retrying_multicall_map(
            koji_session, koji_session.getPackageID, list_of_args=components)
        if not component_ids:
            return cls.compute_weights_from_build_time(components)

        # Prepare list of queries to call koji_session.listBuilds
        build_queries = []
        for component_id in component_ids:
            build_queries.append({
                "packageID": component_id,
                "userID": mbs_user_id,
                "state": koji.BUILD_STATES["COMPLETE"],
                "queryOpts": {"order": "-build_id", "limit": 1}})

        # Get the latest Koji build created by MBS for every component in single Koji call.
        builds_per_component = koji_retrying_multicall_map(
            koji_session, koji_session.listBuilds, list_of_kwargs=build_queries)
        if not builds_per_component:
            return cls.compute_weights_from_build_time(components)

        # Get list of task_ids associated with the latest build in builds.
        # For some packages, there may not be a build done by MBS yet.
        # We store such packages in `components_without_build` and later
        # compute the weight by compute_weights_from_build_time().
        # For others, we will continue by examining weights of all tasks
        # belonging to that build later.
        task_ids = []
        components_with_build = []
        components_without_build = []
        for builds, component_name in zip(builds_per_component, components):
            if not builds:
                # No build for this component.
                components_without_build.append(component_name)
                continue

            latest_build = builds[0]
            task_id = latest_build["task_id"]

            if not task_id:
                # No task_id for this component, this can happen for imported
                # component builds.
                components_without_build.append(component_name)
                continue

            components_with_build.append(component_name)
            task_ids.append(task_id)

        weights = {}

        # For components without any build, fallback to weights computation based on
        # the average time to build.
        weights.update(cls.compute_weights_from_build_time(components_without_build))

        # For components with a build, get the list of tasks associated with this build
        # and compute the weight for each component build as sum of weights of all tasks.
        tasks_per_latest_build = koji_retrying_multicall_map(
            koji_session, koji_session.getTaskDescendents, list_of_args=task_ids)
        if not tasks_per_latest_build:
            return cls.compute_weights_from_build_time(components_with_build)

        for tasks, component_name in zip(tasks_per_latest_build, components_with_build):
            # Compute overall weight of this build. This is sum of weights
            # of all tasks in a build.
            weight = 0
            for task in tasks.values():
                weight += sum([t["weight"] for t in task])
            weights[component_name] = weight

        return weights

    @classmethod
    def get_built_rpms_in_module_build(cls, mmd):
        """
        :param Modulemd mmd: Modulemd to get the built RPMs from.
        :return: list of NVRs
        """
        with models.make_session(conf) as db_session:
            build = models.ModuleBuild.get_build_from_nsvc(
                db_session, mmd.get_name(), mmd.get_stream(), mmd.get_version(),
                mmd.get_context())
            koji_session = KojiModuleBuilder.get_session(conf, login=False)
            rpms = koji_session.listTaggedRPMS(build.koji_tag, latest=True)[0]
            nvrs = set(kobo.rpmlib.make_nvr(rpm, force_epoch=True) for rpm in rpms)
            return list(nvrs)

    def finalize(self):
        # Only import to koji CG if the module is "done".
        if self.config.koji_enable_content_generator and self.module.state == 3:
            cg = KojiContentGenerator(self.module, self.config)
            cg.koji_import()
            if conf.koji_cg_devel_module:
                cg.koji_import(devel=True)

    @staticmethod
    def get_rpm_module_tag(rpm):
        """
        Returns koji tag of a given rpm filename.

        :param str rpm: the *.rpm filename of a rpm
        :rtype: str
        :return: koji tag
        """

        session = KojiModuleBuilder.get_session(conf, login=False)
        rpm_md = session.getRPM(rpm)
        if not rpm_md:
            return None

        tags = []
        koji_tags = session.listTags(rpm_md["build_id"])
        for t in koji_tags:
            if not t["name"].endswith("-build") and t["name"].startswith("module-"):
                tags.append(t["name"])

        return tags
